import asyncio
from typing import List, Optional

from llama_index.core import PromptTemplate
from llama_index.core.prompts import PromptType
from llama_index.core.prompts.utils import is_chat_model
from llama_index.core.response_synthesizers import Refine
from llama_index.core.service_context_elements.llm_predictor import LLMPredictorType

from autorag.nodes.passagecompressor.base import passage_compressor_node
from autorag.utils.util import process_batch


@passage_compressor_node
def refine(queries: List[str],
           contents: List[List[str]],
           scores,
           ids,
           llm: LLMPredictorType,
           prompt: Optional[str] = None,
           chat_prompt: Optional[str] = None,
           batch: int = 16,
           ) -> List[str]:
    """
    Refine a response to a query across text chunks.
    This function is a wrapper for llama_index.response_synthesizers.Refine.
    For more information, visit https://docs.llamaindex.ai/en/stable/examples/response_synthesizers/refine/.

    :param queries: The queries for retrieved passages.
    :param contents: The contents of retrieved passages.
    :param scores: The scores of retrieved passages.
        Do not use in this function, so you can pass an empty list.
    :param ids: The ids of retrieved passages.
        Do not use in this function, so you can pass an empty list.
    :param llm: The llm instance that will be used to summarize.
    :param prompt: The prompt template for refine.
        If you want to use chat prompt, you should pass chat_prompt instead.
        At prompt, you must specify where to put 'context_msg' and 'query_str'.
        Default is None. When it is None, it will use llama index default prompt.
    :param chat_prompt: The chat prompt template for refine.
        If you want to use normal prompt, you should pass prompt instead.
        At prompt, you must specify where to put 'context_msg' and 'query_str'.
        Default is None. When it is None, it will use llama index default chat prompt.
    :param batch: The batch size for llm.
        Set low if you face some errors.
        Default is 16.
    :return: The list of compressed texts.
    """
    if prompt is not None and not is_chat_model(llm):
        refine_template = PromptTemplate(prompt, prompt_type=PromptType.REFINE)
    elif chat_prompt is not None and is_chat_model(llm):
        refine_template = PromptTemplate(chat_prompt, prompt_type=PromptType.REFINE)
    else:
        refine_template = None
    summarizer = Refine(llm=llm,
                        refine_template=refine_template,
                        verbose=True)
    tasks = [summarizer.aget_response(query, content) for query, content in zip(queries, contents)]
    loop = asyncio.get_event_loop()
    results = loop.run_until_complete(process_batch(tasks, batch_size=batch))
    return results



from autorag.nodes.passagereranker import cohere_reranker

from tests.autorag.nodes.passagereranker.test_passage_reranker_base import queries_example, contents_example, \
    scores_example, ids_example, base_reranker_test, project_dir, previous_result, base_reranker_node_test


def test_cohere_reranker():
    top_k = 3
    original_cohere_reranker = cohere_reranker.__wrapped__
    contents_result, id_result, score_result \
        = original_cohere_reranker(queries_example, contents_example, scores_example, ids_example, top_k)
    base_reranker_test(contents_result, id_result, score_result, top_k)


def test_cohere_reranker_batch_one():
    top_k = 3
    batch = 1
    original_cohere_reranker = cohere_reranker.__wrapped__
    contents_result, id_result, score_result \
        = original_cohere_reranker(queries_example, contents_example, scores_example, ids_example, top_k, batch=batch)
    base_reranker_test(contents_result, id_result, score_result, top_k)


def test_cohere_node():
    top_k = 1
    result_df = cohere_reranker(project_dir=project_dir, previous_result=previous_result, top_k=top_k)
    base_reranker_node_test(result_df, top_k)



from datetime import datetime, date

from autorag.nodes.passagefilter import recency_filter
from tests.autorag.nodes.passagefilter.test_passage_filter_base import contents_example, time_list, ids_example, \
    scores_example, base_passage_filter_test, base_passage_filter_node_test, project_dir_with_corpus, \
    previous_result


def test_recency_filter():
    original_recency_filter = recency_filter.__wrapped__
    contents_result, id_result, score_result = original_recency_filter \
        (contents_example, scores_example, ids_example, time_list, threshold=date(2021, 6, 30))
    assert id_result[0] == [ids_example[0][1], ids_example[0][2], ids_example[0][3]]
    assert id_result[1] == [ids_example[1][2], ids_example[1][3]]
    assert contents_result[0] == [contents_example[0][1], contents_example[0][2], contents_example[0][3]]
    assert contents_result[1] == [contents_example[1][2], contents_example[1][3]]
    assert score_result[0] == [0.8, 0.1, 0.5]
    assert score_result[1] == [0.7, 0.3]
    base_passage_filter_test(contents_result, id_result, score_result)


def test_recency_filter_date_time_list():
    time_list_2 = [[date(2015, 1, 1), datetime(2021, 9, 3),
                    datetime(2022, 3, 5, 12, 30),
                    datetime(2022, 3, 5, 12, 45, 00)],
                   [date(2015, 1, 1), date(2021, 1, 1),
                    datetime(2022, 3, 5, 12, 40),
                    datetime(2022, 3, 5, 12, 45, 00)]]
    original_recency_filter = recency_filter.__wrapped__
    contents_result, id_result, score_result = original_recency_filter \
        (contents_example, scores_example, ids_example, time_list_2, threshold=date(2021, 6, 30))
    assert id_result[0] == [ids_example[0][1], ids_example[0][2], ids_example[0][3]]
    assert id_result[1] == [ids_example[1][2], ids_example[1][3]]
    assert contents_result[0] == [contents_example[0][1], contents_example[0][2], contents_example[0][3]]
    assert contents_result[1] == [contents_example[1][2], contents_example[1][3]]
    assert score_result[0] == [0.8, 0.1, 0.5]
    assert score_result[1] == [0.7, 0.3]
    base_passage_filter_test(contents_result, id_result, score_result)


def test_recency_filter_all_filtered():
    original_recency_filter = recency_filter.__wrapped__
    contents_result, id_result, score_result = original_recency_filter \
        (contents_example, scores_example, ids_example, time_list, threshold=datetime(2040, 6, 30))
    assert id_result[0] == [ids_example[0][3]]
    assert id_result[1] == [ids_example[1][3]]
    assert contents_result[0] == [contents_example[0][3]]
    assert contents_result[1] == [contents_example[1][3]]
    assert score_result[0] == [0.5]
    assert score_result[1] == [0.3]
    base_passage_filter_test(contents_result, id_result, score_result)


def test_recency_filter_minutes():
    original_recency_filter = recency_filter.__wrapped__
    contents_result, id_result, score_result = original_recency_filter \
        (contents_example, scores_example, ids_example, time_list,
         threshold=datetime(2022, 3, 5, 12, 35))
    assert id_result[0] == [ids_example[0][3]]
    assert id_result[1] == [ids_example[1][2], ids_example[1][3]]
    assert contents_result[0] == [contents_example[0][3]]
    assert contents_result[1] == [contents_example[1][2], contents_example[1][3]]
    assert score_result[0] == [0.5]
    assert score_result[1] == [0.7, 0.3]
    base_passage_filter_test(contents_result, id_result, score_result)


def test_recency_filter_seconds():
    original_recency_filter = recency_filter.__wrapped__
    contents_result, id_result, score_result = original_recency_filter \
        (contents_example, scores_example, ids_example, time_list,
         threshold=datetime(2022, 3, 5, 12, 40, 30))
    assert id_result[0] == [ids_example[0][3]]
    assert id_result[1] == [ids_example[1][3]]
    assert contents_result[0] == [contents_example[0][3]]
    assert contents_result[1] == [contents_example[1][3]]
    assert score_result[0] == [0.5]
    assert score_result[1] == [0.3]
    base_passage_filter_test(contents_result, id_result, score_result)


def test_recency_filter_node(project_dir_with_corpus):
    result_df = recency_filter(
        project_dir=project_dir_with_corpus, previous_result=previous_result,
        threshold=datetime(2021, 6, 30))
    base_passage_filter_node_test(result_df)



import os
import pathlib
from datetime import datetime

import pandas as pd
import pytest

from autorag.utils import (validate_qa_dataset, validate_corpus_dataset, cast_qa_dataset, cast_corpus_dataset,
                           validate_qa_from_corpus_dataset)


@pytest.fixture
def qa_df():
    return pd.DataFrame({
        'qid': ['id1', 'id2'],
        'query': ['query1', 'query2'],
        'retrieval_gt': [[['doc1', 'doc3'], ['doc2']], [[]]],
        'generation_gt': 'answer1',
    })


# Fixture for corpus dataset
@pytest.fixture
def corpus_df():
    return pd.DataFrame({
        'doc_id': ['doc1', 'doc2', 'doc3'],
        'contents': ['content1', 'content2', 'content3'],
        'metadata': [{'prev_id': None, 'next_id': 'doc2'}, {'test_key': 'test_value'},
                     {'last_modified_datetime': datetime(2022, 12, 1, 3, 4, 5)}]
    })


# Test validate_qa_dataset
def test_validate_qa_dataset(qa_df):
    # This should pass as the qa_df fixture contains the required columns
    validate_qa_dataset(qa_df)

    # This should fail as 'qid' column is missing
    with pytest.raises(AssertionError):
        invalid_df = qa_df.drop(columns=['qid'])
        validate_qa_dataset(invalid_df)


# Test validate_corpus_dataset
def test_validate_corpus_dataset(corpus_df):
    # This should pass as the corpus_df fixture contains the required columns
    validate_corpus_dataset(corpus_df)

    # This should fail as 'doc_id' column is missing
    with pytest.raises(AssertionError):
        invalid_df = corpus_df.drop(columns=['doc_id'])
        validate_corpus_dataset(invalid_df)


# Test cast_qa_dataset
def test_cast_qa_dataset(qa_df):
    # Cast the dataset and check for correct casting
    casted_df = cast_qa_dataset(qa_df)
    assert all(isinstance(x, list) for x in casted_df['retrieval_gt'])
    assert all(isinstance(x[0], list) for x in casted_df['retrieval_gt'])
    assert all(isinstance(x, list) for x in casted_df['generation_gt'])

    # Check for ValueError when 'retrieval_gt' has an incorrect type
    with pytest.raises(ValueError):
        invalid_df = qa_df.copy()
        invalid_df.at[0, 'retrieval_gt'] = 123  # Invalid type for casting
        cast_qa_dataset(invalid_df)


# Test cast_corpus_dataset
def test_cast_corpus_dataset(corpus_df):
    # Cast the dataset and check for a datetime key in metadata
    casted_df = cast_corpus_dataset(corpus_df)
    assert all('last_modified_datetime' in x for x in casted_df['metadata'])
    assert casted_df['metadata'].iloc[0]['prev_id'] is None
    assert casted_df['metadata'].iloc[0]['next_id'] == 'doc2'
    assert casted_df['metadata'].iloc[1]['test_key'] == 'test_value'
    assert casted_df['metadata'].iloc[1]['prev_id'] is None
    assert casted_df['metadata'].iloc[1]['next_id'] is None
    assert casted_df['metadata'].iloc[2]['last_modified_datetime'] == datetime(2022, 12, 1, 3, 4, 5)


def test_validate_qa_from_corpus_dataset(qa_df, corpus_df):
    validate_qa_from_corpus_dataset(qa_df, corpus_df)

    with pytest.raises(AssertionError) as excinfo:
        invalid_df = qa_df.copy()
        invalid_df.at[0, 'retrieval_gt'] = [['answer1', 'answer2'], ['answer3']]
        validate_qa_from_corpus_dataset(invalid_df, corpus_df)
    assert "3 doc_ids in retrieval_gt do not exist in corpus_df." in str(excinfo.value)

    root_dir = pathlib.PurePath(os.path.dirname(os.path.realpath(__file__))).parent.parent
    project_dir = os.path.join(root_dir, "resources", "sample_project")
    qa_parquet = pd.read_parquet(os.path.join(project_dir, "data", "qa.parquet"))

    with pytest.raises(AssertionError) as excinfo:
        validate_qa_from_corpus_dataset(qa_parquet, corpus_df)
    assert "10 doc_ids in retrieval_gt do not exist in corpus_df." in str(excinfo.value)



import os
import pathlib
import pickle
import tempfile
from datetime import datetime

import pandas as pd
import pytest

from autorag.nodes.retrieval import bm25
from autorag.nodes.retrieval.bm25 import bm25_ingest
from tests.autorag.nodes.retrieval.test_retrieval_base import (queries, project_dir, corpus_df, previous_result,
                                                               base_retrieval_test, base_retrieval_node_test)

root_dir = pathlib.PurePath(os.path.dirname(os.path.realpath(__file__))).parent.parent.parent
bm25_path = os.path.join(root_dir, "resources", "test_bm25_retrieval.pkl")
with open(bm25_path, 'rb') as r:
    bm25_corpus = pickle.load(r)


@pytest.fixture
def ingested_bm25_path():
    with tempfile.NamedTemporaryFile(suffix='.pkl', mode='w+b') as path:
        bm25_ingest(path.name, corpus_df)
        yield path.name


def test_bm25_retrieval():
    top_k = 10
    original_bm25 = bm25.__wrapped__
    id_result, score_result = original_bm25(queries, top_k=top_k, bm25_corpus=bm25_corpus)
    base_retrieval_test(id_result, score_result, top_k)


def test_bm25_node():
    result_df = bm25(project_dir=project_dir, previous_result=previous_result, top_k=4)
    base_retrieval_node_test(result_df)


def test_bm25_ingest(ingested_bm25_path):
    with open(ingested_bm25_path, 'rb') as r:
        corpus = pickle.load(r)
    assert set(corpus.keys()) == {'tokens', 'passage_id'}
    assert isinstance(corpus['tokens'], list)
    assert isinstance(corpus['passage_id'], list)
    assert len(corpus['tokens']) == len(corpus['passage_id']) == 5
    assert set(corpus['passage_id']) == {'doc1', 'doc2', 'doc3', 'doc4', 'doc5'}

    bm25_origin = bm25.__wrapped__
    top_k = 2
    id_result, score_result = bm25_origin([['What is test document?'], ['What is test document number 2?']],
                                          top_k=top_k, bm25_corpus=corpus)
    assert len(id_result) == len(score_result) == 2
    for id_list, score_list in zip(id_result, score_result):
        assert isinstance(id_list, list)
        assert isinstance(score_list, list)
        for _id in id_list:
            assert isinstance(_id, str)
            assert _id in ['doc1', 'doc2', 'doc3', 'doc4', 'doc5']


def test_duplicate_id_bm25_ingest(ingested_bm25_path):
    new_doc_id = ["doc4", "doc5", "doc6", "doc7", "doc8"]
    new_contents = ["This is a test document 4.", "This is a test document 5.", "This is a test document 6.",
                    "This is a test document 7.", "This is a test document 8."]
    new_metadata = [{'datetime': datetime.now()} for _ in range(5)]
    new_corpus_df = pd.DataFrame({"doc_id": new_doc_id, "contents": new_contents, "metadata": new_metadata})
    bm25_ingest(ingested_bm25_path, new_corpus_df)
    with open(ingested_bm25_path, 'rb') as r:
        corpus = pickle.load(r)
    assert len(corpus['tokens']) == len(corpus['passage_id']) == 8



import asyncio
import os

from autorag.nodes.passagereranker import jina_reranker
from autorag.nodes.passagereranker.jina import jina_reranker_pure
from tests.autorag.nodes.passagereranker.test_passage_reranker_base import queries_example, contents_example, \
    scores_example, ids_example, base_reranker_test, project_dir, previous_result, base_reranker_node_test


def test_jina_reranker_pure():
    api_key = os.environ['JINAAI_API_KEY']
    assert bool(api_key) is True
    content_result, id_result, score_result = asyncio.run(
        jina_reranker_pure(queries_example[0], contents_example[0], scores_example[0], ids_example[0], top_k=2,
                           api_key=api_key))
    assert len(content_result) == 2
    assert len(id_result) == 2
    assert len(score_result) == 2

    assert all([res in contents_example[0] for res in content_result])
    assert all([res in ids_example[0] for res in id_result])

    # check if the scores are sorted
    assert score_result[0] >= score_result[1]


def test_jina_reranker():
    top_k = 3
    original_jina_reranker = jina_reranker.__wrapped__
    contents_result, id_result, score_result \
        = original_jina_reranker(queries_example, contents_example, scores_example, ids_example, top_k)
    base_reranker_test(contents_result, id_result, score_result, top_k)


def test_jina_reranker_batch_one():
    top_k = 3
    batch = 1
    original_jina_reranker = jina_reranker.__wrapped__
    contents_result, id_result, score_result \
        = original_jina_reranker(queries_example, contents_example, scores_example, ids_example, top_k, batch=batch)
    base_reranker_test(contents_result, id_result, score_result, top_k)


def test_jina_reranker_node():
    top_k = 1
    result_df = jina_reranker(project_dir=project_dir, previous_result=previous_result, top_k=top_k)
    base_reranker_node_test(result_df, top_k)



from autorag.nodes.passageaugmenter import prev_next_augmenter

from tests.autorag.nodes.passageaugmenter.test_base_passage_augmenter import ids_list, project_dir, \
    previous_result, corpus_data, doc_id_list


def test_prev_next_augmenter_next():
    results = prev_next_augmenter.__wrapped__(ids_list, corpus_data, num_passages=1, mode='next')
    assert results == [[doc_id_list[1], doc_id_list[2]],
                       [doc_id_list[3], doc_id_list[4]],
                       [doc_id_list[0], doc_id_list[1], doc_id_list[29]]]


def test_prev_next_augmenter_prev():
    results = prev_next_augmenter.__wrapped__(ids_list, corpus_data, num_passages=1, mode='prev')
    assert results == [[doc_id_list[0], doc_id_list[1]],
                       [doc_id_list[2], doc_id_list[3]],
                       [doc_id_list[0], doc_id_list[28], doc_id_list[29]]]


def test_prev_next_augmenter_both():
    results = prev_next_augmenter.__wrapped__(ids_list, corpus_data, num_passages=1, mode='both')
    assert results == [[doc_id_list[0], doc_id_list[1], doc_id_list[2]],
                       [doc_id_list[2], doc_id_list[3], doc_id_list[4]],
                       [doc_id_list[0], doc_id_list[1], doc_id_list[28], doc_id_list[29]]]


def test_prev_next_augmenter_multi_passages():
    results = prev_next_augmenter.__wrapped__(ids_list, corpus_data, num_passages=3, mode='prev')
    assert results == [[doc_id_list[0], doc_id_list[1]],
                       [doc_id_list[0], doc_id_list[1], doc_id_list[2], doc_id_list[3]],
                       [doc_id_list[0], doc_id_list[26], doc_id_list[27], doc_id_list[28], doc_id_list[29]]]


def test_prev_next_augmenter_node():
    result_df = prev_next_augmenter(project_dir=project_dir, previous_result=previous_result, mode='next', top_k=2)
    contents = result_df["retrieved_contents"].tolist()
    ids = result_df["retrieved_ids"].tolist()
    scores = result_df["retrieve_scores"].tolist()
    assert len(contents) == len(ids) == len(scores) == 2
    assert len(contents[0]) == len(ids[0]) == len(scores[0]) == 2
    for content_list, id_list, score_list in zip(contents, ids, scores):
        for i, (content, _id, score) in enumerate(zip(content_list, id_list, score_list)):
            assert isinstance(content, str)
            assert isinstance(_id, str)
            assert isinstance(score, float)
            assert _id in corpus_data["doc_id"].tolist()
            assert content == corpus_data[corpus_data["doc_id"] == _id]["contents"].values[0]
            if i >= 1:
                assert score_list[i - 1] >= score_list[i]



from autorag.nodes.promptmaker import fstring
from tests.autorag.nodes.promptmaker.test_prompt_maker_base import prompt, queries, retrieved_contents, previous_result


def test_fstring():
    fstring_original = fstring.__wrapped__
    result_prompts = fstring_original(prompt, queries, retrieved_contents)
    assert len(result_prompts) == 2
    assert isinstance(result_prompts, list)
    assert result_prompts[0] == "Answer this question: What is the capital of Japan? \n\n Tokyo is the capital of Japan.\n\nTokyo, the capital of Japan, is a huge metropolitan city."
    assert result_prompts[1] == "Answer this question: What is the capital of China? \n\n Beijing is the capital of China.\n\nBeijing, the capital of China, is a huge metropolitan city."


def test_fstring_node():
    result = fstring(project_dir="pseudo_project_dir",
                     previous_result=previous_result,
                     prompt=prompt)
    assert len(result) == 2
    assert result.columns == ["prompts"]
    assert result['prompts'][0] == "Answer this question: What is the capital of Japan? \n\n Tokyo is the capital of Japan.\n\nTokyo, the capital of Japan, is a huge metropolitan city."
    assert result['prompts'][1] == "Answer this question: What is the capital of China? \n\n Beijing is the capital of China.\n\nBeijing, the capital of China, is a huge metropolitan city."



from typing import List

from autorag.nodes.passagecompressor.base import passage_compressor_node


@passage_compressor_node
def pass_compressor(contents: List[List[str]]):
    """Do not perform any passage compression"""
    return contents



from .base import retrieval_node
from .bm25 import bm25
from .hybrid_cc import hybrid_cc
from .hybrid_dbsf import hybrid_dbsf
from .hybrid_rrf import hybrid_rrf
from .hybrid_rsf import hybrid_rsf
from .vectordb import vectordb



import functools
import time
from typing import List, Iterable, Tuple, Any, Optional, Callable

import pandas as pd


def measure_speed(func, *args, **kwargs):
    """
    Method for measuring execution speed of the function.
    """
    start_time = time.time()
    result = func(*args, **kwargs)
    end_time = time.time()
    return result, end_time - start_time


def avoid_empty_result(return_index: List[int]):
    """
    Decorator for avoiding empty results from the function.
    When the func returns an empty result, it will return the origin results.
    When the func returns a None, it will return the origin results.
    When the return value is a tuple, it will check all the value or list is empty.
    If so, it will return the origin results.
    It keeps parameters at return_index of the function as the origin results.

    :param return_index: The index of the result to be returned when there is no result.
    :return: The origin results or the results from the function.
    """

    def decorator_avoid_empty_result(func: Callable):

        @functools.wraps(func)
        def wrapper(*args, **kwargs) -> List:
            func_result = func(*args, **kwargs)
            if isinstance(func_result, tuple):
                # if all the results are empty, return the origin results.
                if all([not bool(result) for result in func_result]):
                    return [args[index] for index in return_index]
            if not bool(func_result):
                return [args[index] for index in return_index]
            else:
                return func_result

        return wrapper

    return decorator_avoid_empty_result


@avoid_empty_result([0, 3])
def filter_by_threshold(results, value, threshold, metadatas=None) -> Tuple[List, List]:
    """
    Filter results by value's threshold.

    :param results: The result list to be filtered.
    :param value: The value list to be filtered.
        It must have the same length with results.
    :param threshold: The threshold value.
    :param metadatas: The metadata of each result.
    :return: Filtered list of results and filtered list of metadatas.
        Metadatas will be returned even if you did not give input metadatas.
    :rtype: Tuple[List, List]
    """
    if metadatas is None:
        metadatas = [None] * len(results)
    assert len(results) == len(value), "results and value must have the same length."
    try:
        filtered_results, _, filtered_metadatas = zip(
            *filter(lambda x: x[1] <= threshold, zip(results, value, metadatas)))
    except ValueError:
        return [], []
    return list(filtered_results), list(filtered_metadatas)


def select_best_average(results: List[pd.DataFrame], columns: Iterable[str],
                        metadatas: Optional[List[Any]] = None) -> Tuple[pd.DataFrame, Any]:
    """
    Select the best result by average value among given columns.

    :param results: The list of results.
        Each result must be pd.DataFrame.
    :param columns: Column names to be averaged.
        Standard to select the best result.
    :param metadatas: The metadata of each result. 
        It will select one metadata with the best result.
    :return: The best result and the best metadata.
        The metadata will be returned even if you did not give input 'metadatas' parameter.
    :rtype: Tuple[pd.DataFrame, Any]
    """
    if metadatas is None:
        metadatas = [None] * len(results)
    assert len(results) == len(metadatas), "results and module_filename must have the same length."
    assert all([isinstance(result, pd.DataFrame) for result in results]), \
        "results must be pd.DataFrame."
    assert all([column in result.columns for result in results for column in columns]), \
        "columns must be in the columns of results."
    each_average = [df[columns].mean(axis=1).mean() for df in results]
    best_index = each_average.index(max(each_average))
    return results[best_index], metadatas[best_index]



import os.path
import pathlib
import tempfile
from distutils.dir_util import copy_tree
from unittest.mock import patch

import pandas as pd
import pytest
from llama_index.core.base.llms.types import CompletionResponse
from llama_index.llms.openai import OpenAI

from autorag.deploy import extract_best_config
from autorag.evaluator import Evaluator
from autorag.nodes.retrieval import bm25, vectordb, hybrid_rrf
from autorag.nodes.retrieval.run import run_retrieval_node
from autorag.schema import Node
from autorag.utils import validate_qa_dataset, validate_corpus_dataset
from autorag.utils.util import load_summary_file
from tests.delete_tests import is_github_action

root_dir = pathlib.PurePath(os.path.dirname(os.path.realpath(__file__))).parent
resource_dir = os.path.join(root_dir, 'resources')


@pytest.fixture
def evaluator():
    with tempfile.TemporaryDirectory() as project_dir:
        evaluator = Evaluator(os.path.join(resource_dir, 'qa_data_sample.parquet'),
                              os.path.join(resource_dir, 'corpus_data_sample.parquet'),
                              project_dir)
        yield evaluator


@pytest.fixture
def test_evaluator():
    with tempfile.TemporaryDirectory() as project_dir:
        evaluator = Evaluator(os.path.join(resource_dir, 'qa_test_data_sample.parquet'),
                              os.path.join(resource_dir, 'corpus_data_sample.parquet'),
                              project_dir)
        yield evaluator


def test_evaluator_init(evaluator):
    validate_qa_dataset(evaluator.qa_data)
    validate_corpus_dataset(evaluator.corpus_data)
    project_dir = evaluator.project_dir
    loaded_qa_data = pd.read_parquet(os.path.join(project_dir, 'data', 'qa.parquet'))
    loaded_corpus_data = pd.read_parquet(os.path.join(project_dir, 'data', 'corpus.parquet'))
    assert evaluator.qa_data.equals(loaded_qa_data)
    assert evaluator.corpus_data.equals(loaded_corpus_data)


def test_load_node_line(evaluator):
    node_lines = Evaluator._load_node_lines(os.path.join(resource_dir, 'simple.yaml'))
    assert 'retrieve_node_line' in list(node_lines.keys())
    assert node_lines['retrieve_node_line'] is not None
    nodes = node_lines['retrieve_node_line']
    assert isinstance(nodes, list)
    assert len(nodes) == 2
    node = nodes[0]
    assert isinstance(node, Node)
    assert node.node_type == 'retrieval'
    assert node.run_node == run_retrieval_node
    assert node.strategy['metrics'] == ['retrieval_f1', 'retrieval_recall']
    assert node.modules[0].module_type == 'bm25'
    assert node.modules[1].module_type == 'vectordb'
    assert node.modules[2].module_type == 'hybrid_rrf'
    assert node.modules[0].module == bm25
    assert node.modules[1].module == vectordb
    assert node.modules[2].module == hybrid_rrf
    assert node.modules[0].module_param == {}
    assert node.modules[1].module_param == {'embedding_model': ['openai', 'openai']}
    assert node.modules[2].module_param == {
        'rrf_k': 5, 'target_modules': ('bm25', 'vectordb')
    }
    assert nodes[1].node_type == 'passage_filter'


def test_start_trial(evaluator):
    evaluator.start_trial(os.path.join(resource_dir, 'simple.yaml'))
    project_dir = evaluator.project_dir
    assert os.path.exists(os.path.join(project_dir, '0'))
    assert os.path.exists(os.path.join(project_dir, 'data'))
    assert os.path.exists(os.path.join(project_dir, 'resources'))
    assert os.path.exists(os.path.join(project_dir, 'trial.json'))
    assert os.path.exists(os.path.join(project_dir, '0', 'config.yaml'))
    assert os.path.exists(os.path.join(project_dir, '0', 'retrieve_node_line'))
    assert os.path.exists(os.path.join(project_dir, '0', 'retrieve_node_line', 'retrieval'))
    assert os.path.exists(os.path.join(project_dir, '0', 'retrieve_node_line', 'passage_filter'))
    assert os.path.exists(os.path.join(project_dir, '0', 'retrieve_node_line', 'retrieval', '0.parquet'))
    assert os.path.exists(os.path.join(project_dir, '0', 'retrieve_node_line', 'retrieval', '1.parquet'))
    expect_each_result_columns = ['retrieved_contents', 'retrieved_ids', 'retrieve_scores', 'retrieval_f1',
                                  'retrieval_recall']
    each_result_bm25 = pd.read_parquet(
        os.path.join(project_dir, '0', 'retrieve_node_line', 'retrieval', '0.parquet'))
    each_result_vectordb = pd.read_parquet(
        os.path.join(project_dir, '0', 'retrieve_node_line', 'retrieval', '1.parquet'))
    assert all([expect_column in each_result_bm25.columns for expect_column in expect_each_result_columns])
    assert all([expect_column in each_result_vectordb.columns for expect_column in expect_each_result_columns])
    expect_best_result_columns = ['qid', 'query', 'retrieval_gt', 'generation_gt',
                                  'retrieved_contents', 'retrieved_ids', 'retrieve_scores', 'retrieval_f1',
                                  'retrieval_recall']
    best_result = pd.read_parquet(os.path.join(project_dir, '0', 'retrieve_node_line', 'retrieval', 'best_0.parquet'))
    assert all([expect_column in best_result.columns for expect_column in expect_best_result_columns])

    # test node line summary
    node_line_summary_path = os.path.join(project_dir, '0', 'retrieve_node_line', 'summary.csv')
    assert os.path.exists(node_line_summary_path)
    node_line_summary_df = load_summary_file(node_line_summary_path, ["best_module_params"])
    assert len(node_line_summary_df) == 2
    assert set(node_line_summary_df.columns) == {'node_type', 'best_module_filename',
                                                 'best_module_name', 'best_module_params', 'best_execution_time'}
    assert node_line_summary_df['node_type'][0] == 'retrieval'
    assert node_line_summary_df['best_module_filename'][0] == '0.parquet'
    assert node_line_summary_df['best_module_name'][0] == 'bm25'
    assert node_line_summary_df['best_module_params'][0] == {'top_k': 10}
    assert node_line_summary_df['best_execution_time'][0] > 0

    # test trial summary
    trial_summary_path = os.path.join(project_dir, '0', 'summary.csv')
    assert os.path.exists(trial_summary_path)
    trial_summary_df = load_summary_file(trial_summary_path, ["best_module_params"])
    assert len(trial_summary_df) == 2
    assert set(trial_summary_df.columns) == {'node_line_name', 'node_type', 'best_module_filename',
                                             'best_module_name', 'best_module_params', 'best_execution_time'}
    assert trial_summary_df['node_line_name'][0] == 'retrieve_node_line'
    assert trial_summary_df['node_type'][0] == 'retrieval'
    assert trial_summary_df['best_module_filename'][0] == '0.parquet'
    assert trial_summary_df['best_module_name'][0] == 'bm25'
    assert trial_summary_df['best_module_params'][0] == {'top_k': 10}
    assert trial_summary_df['best_execution_time'][0] > 0


@pytest.mark.skip(reason="This test is too slow")
def test_start_trial_full(evaluator):
    evaluator.start_trial(os.path.join(resource_dir, 'full.yaml'))
    project_dir = evaluator.project_dir
    # full path check
    assert os.path.exists(os.path.join(project_dir, '0'))
    assert os.path.exists(os.path.join(project_dir, 'data'))
    assert os.path.exists(os.path.join(project_dir, 'resources'))
    assert os.path.exists(os.path.join(project_dir, 'trial.json'))
    assert os.path.exists(os.path.join(project_dir, '0', 'config.yaml'))

    # node line path check
    # 1. pre_retrieve_node_line
    assert os.path.exists(os.path.join(project_dir, '0', 'pre_retrieve_node_line'))
    assert os.path.exists(os.path.join(project_dir, '0', 'pre_retrieve_node_line', 'query_expansion'))
    assert os.path.exists(os.path.join(project_dir, '0', 'pre_retrieve_node_line', 'query_expansion', "0.parquet"))
    assert os.path.exists(os.path.join(project_dir, '0', 'pre_retrieve_node_line', 'query_expansion', "1.parquet"))
    assert os.path.exists(os.path.join(project_dir, '0', 'pre_retrieve_node_line', 'query_expansion', '2.parquet'))
    assert os.path.exists(os.path.join(project_dir, '0', 'pre_retrieve_node_line', 'query_expansion', '3.parquet'))
    # 2. retrieve_node_line
    retrieval_node_line_path = os.path.join(project_dir, '0', 'retrieve_node_line')
    assert os.path.exists(retrieval_node_line_path)
    retrieval_node_path = os.path.join(retrieval_node_line_path, 'retrieval')
    assert os.path.exists(retrieval_node_path)
    assert all([os.path.exists(os.path.join(retrieval_node_path, f'{i}.parquet')) for i in range(8)])
    assert os.path.exists(os.path.join(project_dir, '0', 'retrieve_node_line', 'passage_reranker'))
    assert os.path.exists(os.path.join(project_dir, '0', 'retrieve_node_line', 'passage_reranker', '0.parquet'))
    assert os.path.exists(os.path.join(project_dir, '0', 'retrieve_node_line', 'passage_reranker', '1.parquet'))
    assert os.path.exists(os.path.join(project_dir, '0', 'retrieve_node_line', 'passage_reranker', '2.parquet'))
    assert os.path.exists(os.path.join(project_dir, '0', 'retrieve_node_line', 'passage_filter'))
    assert os.path.exists(os.path.join(project_dir, '0', 'retrieve_node_line', 'passage_filter', '0.parquet'))
    assert os.path.exists(os.path.join(project_dir, '0', 'retrieve_node_line', 'passage_compressor'))
    assert os.path.exists(os.path.join(project_dir, '0', 'retrieve_node_line', 'passage_compressor', '0.parquet'))
    assert os.path.exists(os.path.join(project_dir, '0', 'retrieve_node_line', 'passage_compressor', '1.parquet'))
    # 3. post_retrieve_node_line
    assert os.path.exists(os.path.join(project_dir, '0', 'post_retrieve_node_line'))
    assert os.path.exists(os.path.join(project_dir, '0', 'post_retrieve_node_line', 'prompt_maker'))
    assert os.path.exists(os.path.join(project_dir, '0', 'post_retrieve_node_line', 'prompt_maker', '0.parquet'))
    assert os.path.exists(os.path.join(project_dir, '0', 'post_retrieve_node_line', 'prompt_maker', '1.parquet'))
    assert os.path.exists(os.path.join(project_dir, '0', 'post_retrieve_node_line', 'generator'))
    assert os.path.exists(os.path.join(project_dir, '0', 'post_retrieve_node_line', 'generator', '0.parquet'))
    assert os.path.exists(os.path.join(project_dir, '0', 'post_retrieve_node_line', 'generator', '1.parquet'))
    assert os.path.exists(os.path.join(project_dir, '0', 'post_retrieve_node_line', 'generator', '2.parquet'))
    assert os.path.exists(os.path.join(project_dir, '0', 'post_retrieve_node_line', 'generator', '3.parquet'))
    assert os.path.exists(os.path.join(project_dir, '0', 'post_retrieve_node_line', 'generator', '4.parquet'))
    assert os.path.exists(os.path.join(project_dir, '0', 'post_retrieve_node_line', 'generator', '5.parquet'))


@pytest.mark.skipif(is_github_action(), reason="Skipping this test on GitHub Actions")
def test_test_data_evaluate(test_evaluator):
    trial_folder = os.path.join(resource_dir, 'result_project', '0')
    with tempfile.NamedTemporaryFile(mode="w+t", suffix=".yaml") as yaml_file:
        extract_best_config(trial_folder, yaml_file.name)
        test_evaluator.start_trial(yaml_file.name)

    project_dir = test_evaluator.project_dir

    assert os.path.exists(os.path.join(project_dir, '0'))
    assert os.path.exists(os.path.join(project_dir, 'data'))
    assert os.path.exists(os.path.join(project_dir, 'resources'))
    assert os.path.exists(os.path.join(project_dir, 'trial.json'))
    assert os.path.exists(os.path.join(project_dir, '0', 'config.yaml'))
    assert os.path.exists(os.path.join(project_dir, '0', 'retrieve_node_line'))
    assert os.path.exists(os.path.join(project_dir, '0', 'retrieve_node_line', 'retrieval'))
    assert os.path.exists(os.path.join(project_dir, '0', 'retrieve_node_line', 'retrieval', '0.parquet'))
    assert os.path.exists(os.path.join(project_dir, '0', 'retrieve_node_line', 'retrieval', 'best_0.parquet'))
    assert os.path.exists(os.path.join(project_dir, '0', 'retrieve_node_line', 'summary.csv'))
    assert os.path.exists(os.path.join(project_dir, '0', 'pre_retrieve_node_line'))
    assert os.path.exists(os.path.join(project_dir, '0', 'pre_retrieve_node_line', 'query_expansion'))
    assert os.path.exists(os.path.join(project_dir, '0', 'pre_retrieve_node_line', 'query_expansion', "0.parquet"))
    assert os.path.exists(os.path.join(project_dir, '0', 'post_retrieve_node_line'))
    assert os.path.exists(os.path.join(project_dir, '0', 'post_retrieve_node_line', 'prompt_maker'))
    assert os.path.exists(os.path.join(project_dir, '0', 'post_retrieve_node_line', 'generator'))
    assert os.path.exists(os.path.join(project_dir, '0', 'summary.csv'))


def base_restart_trial(evaluator, error_folder_path):
    error_path = os.path.join(evaluator.project_dir, '0')
    copy_tree(error_folder_path, error_path)
    evaluator.restart_trial(error_path)
    assert os.path.exists(os.path.join(error_path, 'summary.csv'))
    assert os.path.exists(os.path.join(error_path, 'pre_retrieve_node_line'))
    assert os.path.exists(os.path.join(error_path, 'pre_retrieve_node_line', 'summary.csv'))
    assert os.path.exists(os.path.join(error_path, 'retrieve_node_line'))
    assert os.path.exists(os.path.join(error_path, 'retrieve_node_line', 'summary.csv'))
    assert os.path.exists(os.path.join(error_path, 'post_retrieve_node_line'))
    assert os.path.exists(os.path.join(error_path, 'post_retrieve_node_line', 'summary.csv'))


async def mock_acomplete(self, messages, **kwargs):
    return CompletionResponse(text=messages)


async def mock_apredict(self, prompt, **kwargs):
    return prompt.format(**kwargs)


@patch.object(OpenAI, "acomplete", mock_acomplete)
@patch.object(OpenAI, "apredict", mock_apredict)
def test_restart_last_node(evaluator):
    compressor_error_folder_path = os.path.join(resource_dir, 'result_project', '1')
    base_restart_trial(evaluator, compressor_error_folder_path)


@patch.object(OpenAI, "acomplete", mock_acomplete)
def test_restart_first_node(evaluator):
    prompt_error_folder_path = os.path.join(resource_dir, 'result_project', '2')
    base_restart_trial(evaluator, prompt_error_folder_path)


def test_restart_leads_start_trial(evaluator):
    start_error_folder_path = os.path.join(resource_dir, 'result_project')
    copy_tree(start_error_folder_path, evaluator.project_dir)
    error_path = os.path.join(evaluator.project_dir, '3')
    evaluator.restart_trial(error_path)
    restart_path = os.path.join(evaluator.project_dir, '4')
    assert os.path.exists(os.path.join(restart_path, 'summary.csv'))
    assert os.path.exists(os.path.join(restart_path, 'retrieve_node_line'))
    assert os.path.exists(os.path.join(restart_path, 'retrieve_node_line', 'summary.csv'))



import asyncio
from typing import List, Tuple

import torch
from transformers import T5Tokenizer, T5ForConditionalGeneration

from autorag.nodes.passagereranker.base import passage_reranker_node


@passage_reranker_node
def upr(queries: List[str], contents_list: List[List[str]],
        scores_list: List[List[float]], ids_list: List[List[str]],
        top_k: int, shard_size: int = 16, use_bf16: bool = False,
        prefix_prompt: str = "Passage: ",
        suffix_prompt: str = "Please write a question based on this passage.") \
        -> Tuple[List[List[str]], List[List[str]], List[List[float]]]:
    """
    Rerank a list of contents based on their relevance to a query using UPR.
    UPR is a reranker based on UPR (https://github.com/DevSinghSachan/unsupervised-passage-reranking).
    The language model will make a question based on the passage and rerank the passages by the likelihood of the question.
    The default model is t5-large.

    :param queries: The list of queries to use for reranking
    :param contents_list: The list of lists of contents to rerank
    :param scores_list: The list of lists of scores retrieved from the initial ranking
    :param ids_list: The list of lists of ids retrieved from the initial ranking
    :param top_k: The number of passages to be retrieved
    :param shard_size: The shard size for the model.
        The larger the shard size, the faster the reranking speed.
        But it will consume more memory and compute power.
        Default is 16.
    :param use_bf16: Whether to use bfloat16 for the model. Default is False.
    :param prefix_prompt: The prefix prompt for the language model that generates question for reranking.
        Default is "Passage: ".
        The prefix prompt serves as the initial context or instruction for the language model.
        It sets the stage for what is expected in the output
    :param suffix_prompt: The suffix prompt for the language model that generates question for reranking.
        Default is "Please write a question based on this passage.".
        The suffix prompt provides a cue or a closing instruction to the language model,
            signaling how to conclude the generated text or what format to follow at the end.
    :return: tuple of lists containing the reranked contents, ids, and scores
    """
    # Load the tokenizer and model
    model_name = "t5-large"
    tokenizer = T5Tokenizer.from_pretrained(model_name)
    model = T5ForConditionalGeneration.from_pretrained(model_name,
                                                       torch_dtype=torch.bfloat16 if use_bf16 else torch.float32)
    # Determine the device to run the model on (GPU if available, otherwise CPU)
    device = ("cuda" if torch.cuda.is_available() else "cpu")
    # Run async upr_rerank_pure function
    tasks = [upr_pure(query, contents, scores,
                      ids, top_k, model, device, tokenizer,
                      shard_size, prefix_prompt, suffix_prompt)
             for query, contents, scores, ids in
             zip(queries, contents_list, scores_list, ids_list)]
    loop = asyncio.get_event_loop()
    results = loop.run_until_complete(asyncio.gather(*tasks))

    content_result = list(map(lambda x: x[0], results))
    id_result = list(map(lambda x: x[1], results))
    score_result = list(map(lambda x: x[2], results))

    del model
    del tokenizer
    if torch.cuda.is_available():
        torch.cuda.empty_cache()

    return content_result, id_result, score_result


async def upr_pure(query: str, contents: List[str], scores: List[float],
                   ids: List[str], top_k: int, model, device, tokenizer,
                   shard_size: int, prefix_prompt: str, suffix_prompt: str) \
        -> Tuple[List[str], List[str], List[float]]:
    """
    Rerank a list of contents based on their relevance to a query using UPR.

    :param query: The query to use for reranking
    :param contents: The list of contents to rerank
    :param scores: The list of scores retrieved from the initial ranking
    :param ids: The list of ids retrieved from the initial ranking
    :param top_k: The number of passages to be retrieved
    :param model: The UPR model to use for reranking
    :param device: The device to run the model on (GPU if available, otherwise CPU)
    :param tokenizer: The tokenizer to use for the model
    :param shard_size: The shard size for the model.
    :param prefix_prompt: The prefix prompt for the language model that generates question for reranking.
    :param suffix_prompt: The suffix prompt for the language model that generates question for reranking.
    :return: tuple of lists containing the reranked contents, ids, and scores
    """
    indexes, scores = calculate_likelihood(query, contents, prefix_prompt, suffix_prompt,
                                           tokenizer, device, model, shard_size)
    reranked_contents, reranked_ids = zip(*[(contents[idx], ids[idx]) for idx in indexes])

    # crop with top_k
    if len(reranked_contents) < top_k:
        top_k = len(reranked_contents)
    reranked_contents, reranked_ids, scores = reranked_contents[:top_k], reranked_ids[:top_k], scores[:top_k]

    return list(reranked_contents), list(reranked_ids), list(scores)


def calculate_likelihood(query: str, contents: List[str],
                         prefix_prompt: str, suffix_prompt: str,
                         tokenizer, device, model, shard_size: int)\
        -> tuple[List[int], List[float]]:
    # create prompts
    prompts = [f"{prefix_prompt} {content} {suffix_prompt}" for content in contents]

    # tokenize contexts and instruction prompts
    context_tokens = tokenizer(prompts,
                               padding='longest',
                               max_length=512,
                               pad_to_multiple_of=8,
                               truncation=True,
                               return_tensors='pt')
    context_tensor, context_attention_mask = context_tokens.input_ids, context_tokens.attention_mask
    if device == 'cuda':
        context_tensor, context_attention_mask = context_tensor.cuda(), context_attention_mask.cuda()

    # tokenize question
    question_tokens = tokenizer([query],
                                max_length=128,
                                truncation=True,
                                return_tensors='pt')
    question_tensor = question_tokens.input_ids
    if device == 'cuda':
        question_tensor = question_tensor.cuda()
    question_tensor = torch.repeat_interleave(question_tensor, len(contents), dim=0)

    if device == 'cuda':
        model = model.to(device)

    sharded_nll_list = []

    # calculate log likelihood
    for i in range(0, len(context_tensor), shard_size):
        encoder_tensor_view = context_tensor[i: i + shard_size]
        attention_mask_view = context_attention_mask[i: i + shard_size]
        decoder_tensor_view = question_tensor[i: i + shard_size]
        with torch.no_grad():
            logits = model(input_ids=encoder_tensor_view,
                           attention_mask=attention_mask_view,
                           labels=decoder_tensor_view).logits

        log_softmax = torch.nn.functional.log_softmax(logits, dim=-1)
        nll = -log_softmax.gather(2, decoder_tensor_view.unsqueeze(2)).squeeze(2)

        avg_nll = torch.sum(nll, dim=1)
        sharded_nll_list.append(avg_nll)

    topk_scores, indexes = torch.topk(-torch.cat(sharded_nll_list), k=len(context_tensor))

    return indexes.tolist(), topk_scores.tolist()



import functools
import warnings
from typing import List, Callable, Any, Tuple

import pandas as pd

from autorag.evaluate.metric import retrieval_recall, retrieval_precision, retrieval_f1


def evaluate_retrieval(retrieval_gt: List[List[List[str]]], metrics: List[str]):
    def decorator_evaluate_retrieval(
            func: Callable[[Any], Tuple[List[List[str]], List[List[str]], List[List[float]]]]):
        """
        Decorator for evaluating retrieval results.
        You can use this decorator to any method that returns (contents, scores, ids),
        which is the output of conventional retrieval modules.

        :param func: Must return (contents, scores, ids)
        :return: wrapper function that returns pd.DataFrame, which is the evaluation result.
        """

        @functools.wraps(func)
        def wrapper(*args, **kwargs) -> pd.DataFrame:
            contents, pred_ids, scores = func(*args, **kwargs)
            metric_funcs = {
                retrieval_recall.__name__: retrieval_recall,
                retrieval_precision.__name__: retrieval_precision,
                retrieval_f1.__name__: retrieval_f1,
            }

            metric_scores = {}
            for metric in metrics:
                if metric not in metric_funcs:
                    warnings.warn(f"metric {metric} is not in supported metrics: {metric_funcs.keys()}"
                                  f"{metric} will be ignored.")
                else:
                    metric_func = metric_funcs[metric]
                    metric_scores[metric] = metric_func(retrieval_gt=retrieval_gt, pred_ids=pred_ids)

            metric_result_df = pd.DataFrame(metric_scores)
            execution_result_df = pd.DataFrame({
                'retrieved_contents': contents,
                'retrieved_ids': pred_ids,
                'retrieve_scores': scores,
            })
            result_df = pd.concat([execution_result_df, metric_result_df], axis=1)
            return result_df

        return wrapper

    return decorator_evaluate_retrieval



import pytest

from autorag.nodes.retrieval import hybrid_rsf
from tests.autorag.nodes.retrieval.test_hybrid_base import (sample_ids, sample_scores, base_hybrid_weights_node_test,
                                                            pseudo_project_dir)


def test_hybrid_rsf():
    result_id, result_scores = hybrid_rsf.__wrapped__(sample_ids, sample_scores, top_k=3)
    assert result_id == [
        ['id-3', 'id-1', 'id-2'],
        ['id-4', 'id-2', 'id-3']
    ]
    assert result_scores[0] == pytest.approx([1.0, 0.25, 0.25])
    assert result_scores[1] == pytest.approx([1.0, 0.25, 0.25])


def test_hybrid_rsf_node(pseudo_project_dir):
    retrieve_scores = [1.0, 0.23792372, 0.175]
    base_hybrid_weights_node_test(hybrid_rsf, pseudo_project_dir, retrieve_scores)



from typing import List, Tuple

import pandas as pd
import torch
from FlagEmbedding import FlagReranker

from autorag.nodes.passagereranker.base import passage_reranker_node
from autorag.utils.util import make_batch, sort_by_scores, flatten_apply, select_top_k


@passage_reranker_node
def flag_embedding_reranker(queries: List[str], contents_list: List[List[str]],
                            scores_list: List[List[float]], ids_list: List[List[str]],
                            top_k: int, batch: int = 64, use_fp16: bool = False,
                            model_name: str = "BAAI/bge-reranker-large",
                            ) -> Tuple[List[List[str]], List[List[str]], List[List[float]]]:
    """
    Rerank a list of contents based on their relevance to a query using BAAI normal-Reranker model.

    :param queries: The list of queries to use for reranking
    :param contents_list: The list of lists of contents to rerank
    :param scores_list: The list of lists of scores retrieved from the initial ranking
    :param ids_list: The list of lists of ids retrieved from the initial ranking
    :param top_k: The number of passages to be retrieved
    :param batch: The number of queries to be processed in a batch
        Default is 64.
    :param use_fp16: Whether to use fp16 for inference
    :param model_name: The name of the BAAI Reranker normal-model name.
        Default is "BAAI/bge-reranker-large"
    :return: tuple of lists containing the reranked contents, ids, and scores
    """

    model = FlagReranker(
        model_name_or_path=model_name, use_fp16=use_fp16
    )
    nested_list = [list(map(lambda x: [query, x], content_list)) for query, content_list in zip(queries, contents_list)]
    rerank_scores = flatten_apply(flag_embedding_run_model, nested_list, model=model, batch_size=batch)

    df = pd.DataFrame({
        'contents': contents_list,
        'ids': ids_list,
        'scores': rerank_scores,
    })
    df[['contents', 'ids', 'scores']] = df.apply(sort_by_scores, axis=1, result_type='expand')
    results = select_top_k(df, ['contents', 'ids', 'scores'], top_k)

    del model
    if torch.cuda.is_available():
        torch.cuda.empty_cache()

    return results['contents'].tolist(), results['ids'].tolist(), results['scores'].tolist()


def flag_embedding_run_model(input_texts, model, batch_size: int):
    batch_input_texts = make_batch(input_texts, batch_size)
    results = []
    for batch_texts in batch_input_texts:
        with torch.no_grad():
            pred_scores = model.compute_score(sentence_pairs=batch_texts)
        if batch_size == 1:
            results.append(pred_scores)
        else:
            results.extend(pred_scores)
    return results



from autorag.nodes.passagefilter import pass_passage_filter
from tests.autorag.nodes.passagefilter.test_passage_filter_base import project_dir, previous_result, contents_example, \
    ids_example, scores_example


def test_pass_passage_filter():
    result_df = pass_passage_filter(
        project_dir=project_dir,
        previous_result=previous_result,
    )
    assert all([column_name in result_df.columns for column_name in
                ["retrieved_contents", "retrieved_ids", "retrieve_scores"]])
    assert result_df["retrieved_contents"].tolist() == contents_example
    assert result_df["retrieved_ids"].tolist() == ids_example
    assert result_df["retrieve_scores"].tolist() == scores_example



import tempfile

import pytest
from llama_index.core import Document
from llama_index.core.schema import TextNode, NodeRelationship, RelatedNodeInfo

from autorag.data.corpus import llama_documents_to_parquet, llama_text_node_to_parquet
from tests.autorag.data.corpus.test_base import validate_corpus


@pytest.fixture
def parquet_filepath():
    with tempfile.NamedTemporaryFile(suffix='.parquet') as temp_file:
        yield temp_file.name


def test_llama_documents_to_parquet(parquet_filepath):
    documents = [Document(text='test text', metadata={'key': 'value'},
                          ) for _ in range(5)]
    result_df = llama_documents_to_parquet(documents, parquet_filepath, upsert=True)
    validate_corpus(result_df, 5, parquet_filepath)


def test_llama_text_node_to_parquet(parquet_filepath):
    sample_text_node = TextNode(text='test text',
                                metadata={'key': 'value'},
                                relationships={NodeRelationship.PREVIOUS: RelatedNodeInfo(node_id='0'),
                                               NodeRelationship.NEXT: RelatedNodeInfo(node_id='2')})
    text_nodes = [sample_text_node for _ in range(5)]
    result_df = llama_text_node_to_parquet(text_nodes, parquet_filepath, upsert=True)
    validate_corpus(result_df, 5, parquet_filepath)



import functools
import warnings
from typing import List, Callable, Union, Dict

import pandas as pd

from autorag.evaluate.metric.generation import bleu, meteor, rouge, sem_score, g_eval, bert_score
from autorag.evaluate.util import cast_metrics

GENERATION_METRIC_FUNC_DICT = {func.__name__: func for func in
                               [bleu, meteor, rouge, sem_score, g_eval, bert_score]}


def evaluate_generation(generation_gt: List[List[str]], metrics: Union[List[str], List[Dict]]):
    def decorator_evaluate_generation(func: Callable):
        @functools.wraps(func)
        def wrapper(*args, **kwargs) -> pd.DataFrame:
            generation_result = func(*args, **kwargs)
            if type(generation_result) is tuple:
                assert type(generation_result[0]) is list and type(generation_result[0][0]) is str, \
                    "Input func must return string list as generated answer at the first return value."
                generated_str = generation_result[0]
            elif type(generation_result) is list:
                assert type(generation_result[0]) is str, \
                    "Input func must return string list as generated answer at the first return value."
                generated_str = generation_result
            else:
                raise ValueError("Input func must return string list as generated answer at the first return value.")

            metric_scores = {}
            metric_names, metric_params = cast_metrics(metrics)

            for metric_name, metric_param in zip(metric_names, metric_params):
                if metric_name not in GENERATION_METRIC_FUNC_DICT:
                    warnings.warn(f"metric {metric_name} is not in supported metrics: {GENERATION_METRIC_FUNC_DICT.keys()}"
                                  f"{metric_name} will be ignored.")
                else:
                    metric_scores[metric_name] = GENERATION_METRIC_FUNC_DICT[metric_name](
                        generation_gt=generation_gt, generations=generated_str, **metric_param)

            metric_result_df = pd.DataFrame(metric_scores)
            execution_result_df = pd.DataFrame({
                'generated_texts': generated_str
            })
            if type(generation_result) is tuple:
                execution_result_df['generated_tokens'] = generation_result[1]
                execution_result_df['generated_log_probs'] = generation_result[2]

            result_df = pd.concat([execution_result_df, metric_result_df], axis=1)
            return result_df

        return wrapper

    return decorator_evaluate_generation



import os
import tempfile

import pandas as pd
import pytest

from autorag.nodes.passagefilter import similarity_threshold_cutoff
from autorag.nodes.passagefilter.run import run_passage_filter_node
from autorag.utils.util import load_summary_file
from tests.autorag.nodes.passagefilter.test_passage_filter_base import qa_data, corpus_data, previous_result


@pytest.fixture
def node_line_dir():
    with tempfile.TemporaryDirectory() as project_dir:
        data_dir = os.path.join(project_dir, "data")
        os.makedirs(data_dir)
        qa_data.to_parquet(os.path.join(data_dir, "qa.parquet"), index=False)
        corpus_data.to_parquet(os.path.join(data_dir, "corpus.parquet"), index=False)
        trial_dir = os.path.join(project_dir, "0")
        os.makedirs(trial_dir)
        node_line_dir = os.path.join(trial_dir, "node_line_1")
        os.makedirs(node_line_dir)
        yield node_line_dir


def test_run_passage_filter_node(node_line_dir):
    modules = [similarity_threshold_cutoff]
    module_params = [{'threshold': 0.87}]
    strategies = {
        'metrics': ['retrieval_f1', 'retrieval_recall'],
    }
    best_result = run_passage_filter_node(modules, module_params, previous_result, node_line_dir, strategies)
    assert os.path.exists(os.path.join(node_line_dir, "passage_filter"))
    assert set(best_result.columns) == {'qid', 'query', 'retrieval_gt', 'generation_gt',
                                        'retrieved_contents', 'retrieved_ids', 'retrieve_scores',
                                        'retrieval_f1', 'retrieval_recall',
                                        'passage_filter_retrieval_f1',
                                        'passage_filter_retrieval_recall'}
    # test summary feature
    summary_path = os.path.join(node_line_dir, "passage_filter", "summary.csv")
    assert os.path.exists(summary_path)
    result_path = os.path.join(node_line_dir, "passage_filter", '0.parquet')
    assert os.path.exists(result_path)
    result_df = pd.read_parquet(result_path)
    summary_df = load_summary_file(summary_path)
    assert set(summary_df.columns) == {'filename', 'passage_filter_retrieval_f1',
                                       'passage_filter_retrieval_recall',
                                       'module_name', 'module_params', 'execution_time', 'is_best'}
    assert len(summary_df) == 1
    assert summary_df['filename'][0] == "0.parquet"
    assert summary_df['passage_filter_retrieval_f1'][0] == result_df['retrieval_f1'].mean()
    assert summary_df['passage_filter_retrieval_recall'][0] == result_df['retrieval_recall'].mean()
    assert summary_df['module_name'][0] == "similarity_threshold_cutoff"
    assert summary_df['module_params'][0] == {'threshold': 0.87}
    assert summary_df['execution_time'][0] > 0
    # test the best file is saved properly
    best_path = summary_df[summary_df['is_best']]['filename'].values[0]
    assert os.path.exists(os.path.join(node_line_dir, "passage_filter", f"best_{best_path}"))



from typing import List

from autorag.nodes.passageaugmenter.base import passage_augmenter_node


@passage_augmenter_node
def pass_passage_augmenter(ids_list: List[List[str]], contents_list: List[List[str]], scores_list: List[List[float]]):
    """
    Do not perform augmentation.
    Return given passages, scores, and ids as is.
    """
    return ids_list, contents_list, scores_list



import logging
import os
import pathlib
from typing import List, Callable, Dict

import pandas as pd

from autorag.nodes.retrieval.run import evaluate_retrieval_node
from autorag.strategy import measure_speed, filter_by_threshold, select_best_average

logger = logging.getLogger("AutoRAG")


def run_passage_augmenter_node(modules: List[Callable],
                               module_params: List[Dict],
                               previous_result: pd.DataFrame,
                               node_line_dir: str,
                               strategies: Dict,
                               ) -> pd.DataFrame:
    if not os.path.exists(node_line_dir):
        os.makedirs(node_line_dir)
    project_dir = pathlib.PurePath(node_line_dir).parent.parent
    retrieval_gt = pd.read_parquet(os.path.join(project_dir, "data", "qa.parquet"))['retrieval_gt'].tolist()
    retrieval_gt = [[[str(uuid) for uuid in sub_array] if sub_array.size > 0 else [] for sub_array in inner_array]
                    for inner_array in retrieval_gt]

    results, execution_times = zip(*map(lambda task: measure_speed(
        task[0], project_dir=project_dir, previous_result=previous_result, **task[1]), zip(modules, module_params)))
    average_times = list(map(lambda x: x / len(results[0]), execution_times))

    # run metrics before filtering
    if strategies.get('metrics') is None:
        raise ValueError("You must at least one metrics for passage_augmenter evaluation.")
    results = list(map(lambda x: evaluate_retrieval_node(x, retrieval_gt, strategies.get('metrics')), results))

    # save results to folder
    save_dir = os.path.join(node_line_dir, "passage_augmenter")  # node name
    if not os.path.exists(save_dir):
        os.makedirs(save_dir)
    filepaths = list(map(lambda x: os.path.join(save_dir, f'{x}.parquet'), range(len(modules))))
    list(map(lambda x: x[0].to_parquet(x[1], index=False), zip(results, filepaths)))  # execute save to parquet
    filenames = list(map(lambda x: os.path.basename(x), filepaths))

    summary_df = pd.DataFrame({
        'filename': filenames,
        'module_name': list(map(lambda module: module.__name__, modules)),
        'module_params': module_params,
        'execution_time': average_times,
        **{f'passage_augmenter_{metric}': list(map(lambda result: result[metric].mean(), results)) for metric in
           strategies.get('metrics')},
    })

    # filter by strategies
    if strategies.get('speed_threshold') is not None:
        results, filenames = filter_by_threshold(results, average_times, strategies['speed_threshold'], filenames)
    selected_result, selected_filename = select_best_average(results, strategies.get('metrics'), filenames)
    # change metric name columns to passage_augmenter_metric_name
    selected_result = selected_result.rename(columns={
        metric_name: f'passage_augmenter_{metric_name}' for metric_name in strategies['metrics']})
    # drop retrieval result columns in previous_result
    previous_result = previous_result.drop(columns=['retrieved_contents', 'retrieved_ids', 'retrieve_scores'])
    best_result = pd.concat([previous_result, selected_result], axis=1)

    # add 'is_best' column to summary file
    summary_df['is_best'] = summary_df['filename'] == selected_filename

    # save files
    summary_df.to_csv(os.path.join(save_dir, "summary.csv"), index=False)
    best_result.to_parquet(os.path.join(save_dir, f'best_{os.path.splitext(selected_filename)[0]}.parquet'),
                           index=False)
    return best_result



import logging
from typing import List

from autorag.nodes.promptmaker.base import prompt_maker_node

logger = logging.getLogger("AutoRAG")


@prompt_maker_node
def long_context_reorder(prompt: str, queries: List[str], retrieved_contents: List[List[str]],
                         retrieve_scores: List[List[float]]) -> List[str]:
    """
    Models struggle to access significant details found
    in the center of extended contexts. A study
    (https://arxiv.org/abs/2307.03172) observed that the best
    performance typically arises when crucial data is positioned
    at the start or conclusion of the input context. Additionally,
    as the input context lengthens, performance drops notably, even
    in models designed for long contexts.".

    .. Code:: yaml
    nodes:
    - node_type: prompt_maker
      modules:
      - module_type: long_context_reorder
        prompt: [Answer this question: {query} \n\n {retrieved_contents},
        Read the passages carefully and answer this question: {query} \n\n Passages: {retrieved_contents}]

    :param prompt: A prompt string.
    :param queries: List of query strings.
    :param retrieved_contents: List of retrieved contents.
    :param retrieve_scores: List of retrieve scores.
    :return: Prompts that made by long context reorder.
    """

    def long_context_reorder_row(_prompt: str, _query: str, _retrieved_contents: List[str],
                                 _retrieve_scores: List[float]) -> str:
        if not len(_retrieved_contents) == len(_retrieve_scores):
            logger.info("If you use a summarizer, the reorder will not proceed.")
            return _prompt.format(query=_query, retrieved_contents="\n\n".join(_retrieved_contents))
        content_scores = list(zip(_retrieved_contents, _retrieve_scores))
        sorted_content_scores = sorted(content_scores, key=lambda x: x[1], reverse=True)
        content_result, score_result = zip(*sorted_content_scores)
        _retrieved_contents.append(content_result[0])
        contents_str = "\n\n".join(_retrieved_contents)
        return _prompt.format(query=_query, retrieved_contents=contents_str)

    return list(map(lambda x: long_context_reorder_row(prompt, x[0], x[1], x[2]),
                    zip(queries, retrieved_contents, retrieve_scores)))



import os
import tempfile

import pandas as pd
import pytest

from autorag import generator_models
from autorag.nodes.generator import llama_index_llm
from autorag.nodes.generator.run import run_generator_node
from autorag.utils.util import load_summary_file
from tests.mock import MockLLM

qa_df = pd.DataFrame({
    'qid': ['id-1', 'id-2', 'id-3'],
    'query': ['query-1', 'query-2', 'query-3'],
    'generation_gt': [
        ['The dog had bit the man.', 'The man had bitten the dog.'],
        ['I want to be a artist, but I end up to be a programmer.'],
        ['To be a artist these days, you can overcome by AI.',
         'To be a programmer these days, you can overcome by AI.',
         'To be a lawyer these days, you can overcome by AI.'],
    ],
})

previous_df = pd.DataFrame({
    'qid': ['id-1', 'id-2', 'id-3'],
    'query': ['query-1', 'query-2', 'query-3'],
    'prompts': [
        'What was the dog doing with the man?',
        'What is your dream job? And what is your current job?',
        'Is AI can overcome and replace all jobs in the future?',
    ],
})


@pytest.fixture
def node_line_dir():
    with tempfile.TemporaryDirectory() as temp_dir:
        os.makedirs(os.path.join(temp_dir, "data"))
        qa_df.to_parquet(os.path.join(temp_dir, "data", "qa.parquet"), index=False)
        trial_dir = os.path.join(temp_dir, "test_trial")
        os.makedirs(trial_dir)
        node_line_dir = os.path.join(trial_dir, "test_node_line")
        yield node_line_dir


def test_run_generator_node(node_line_dir):
    generator_models['mock'] = MockLLM
    modules = [llama_index_llm, llama_index_llm]
    module_params = [{'llm': 'mock', 'temperature': 0.5, 'top_p': 0.9, 'max_tokens': 128, 'batch': 8},
                     {'llm': 'mock', 'temperature': 1.5, 'top_p': 0.9, 'max_tokens': 32, 'batch': 8}]
    strategies = {
        'metrics': [{'metric_name': 'bleu'}, {'metric_name': 'meteor'}, {'metric_name': 'rouge'}],
        'speed_threshold': 5,
        'token_threshold': 64,
    }
    best_result = run_generator_node(modules, module_params, previous_df, node_line_dir, strategies)
    assert os.path.exists(os.path.join(node_line_dir, "generator"))
    expect_columns = {'qid', 'query', 'prompts', 'generated_texts', 'generated_tokens', 'generated_log_probs',
                      'bleu', 'meteor', 'rouge'}
    assert set(best_result.columns) == expect_columns

    summary_path = os.path.join(node_line_dir, "generator", "summary.csv")
    assert os.path.exists(summary_path)
    summary_df = load_summary_file(summary_path)
    expect_columns = {'filename', 'bleu', 'meteor', 'rouge', 'module_name', 'module_params', 'execution_time',
                      'average_output_token', 'is_best'}
    assert set(summary_df.columns) == expect_columns
    assert len(summary_df) == 2
    assert summary_df['module_params'][0] == {'llm': 'mock', 'temperature': 0.5, 'top_p': 0.9, 'max_tokens': 128,
                                              'batch': 8}

    first_path = os.path.join(node_line_dir, "generator", "0.parquet")
    assert os.path.exists(first_path)

    best_path = summary_df[summary_df['is_best']]['filename'].values[0]
    assert os.path.exists(os.path.join(node_line_dir, "generator", f'best_{best_path}'))



import itertools
import os.path
import pathlib
from typing import Callable, List, Dict

import pandas as pd

from autorag.evaluate.metric import retrieval_token_recall, retrieval_token_precision, retrieval_token_f1
from autorag.strategy import measure_speed, filter_by_threshold, select_best_average
from autorag.utils import validate_qa_dataset, validate_corpus_dataset
from autorag.utils.util import fetch_contents


def run_passage_compressor_node(modules: List[Callable],
                                module_params: List[Dict],
                                previous_result: pd.DataFrame,
                                node_line_dir: str,
                                strategies: Dict,
                                ) -> pd.DataFrame:
    """
    Run evaluation and select the best module among passage compressor modules.

    :param modules: Passage compressor modules to run.
    :param module_params: Passage compressor module parameters.
    :param previous_result: Previous result dataframe.
        Could be retrieval, reranker modules result.
        It means it must contain 'query', 'retrieved_contents', 'retrieved_ids', 'retrieve_scores' columns.
    :param node_line_dir: This node line's directory.
    :param strategies: Strategies for passage compressor node.
        In this node, we use
        You can skip evaluation when you use only one module and a module parameter.
    :return: The best result dataframe with previous result columns.
        This node will replace 'retrieved_contents' to compressed passages, so its length will be one.
    """
    if not os.path.exists(node_line_dir):
        os.makedirs(node_line_dir)
    project_dir = pathlib.PurePath(node_line_dir).parent.parent
    data_dir = os.path.join(project_dir, "data")
    save_dir = os.path.join(node_line_dir, "passage_compressor")
    if not os.path.exists(save_dir):
        os.makedirs(save_dir)

    # run modules
    results, execution_times = zip(*map(lambda task: measure_speed(
        task[0], project_dir=project_dir, previous_result=previous_result, **task[1]), zip(modules, module_params)))
    results = list(results)
    average_times = list(map(lambda x: x / len(results[0]), execution_times))

    # make retrieval contents gt
    qa_data = pd.read_parquet(os.path.join(data_dir, "qa.parquet"))
    corpus_data = pd.read_parquet(os.path.join(data_dir, "corpus.parquet"))
    retrieval_contents_gt = make_contents_gt(qa_data, corpus_data)

    # run metrics before filtering
    if strategies.get('metrics') is None:
        raise ValueError("You must at least one metrics for retrieval contents evaluation."
                         "It can be 'retrieval_token_f1', 'retrieval_token_precision', 'retrieval_token_recall'.")
    results = list(map(lambda x: evaluate_passage_compressor_node(x, retrieval_contents_gt,
                                                                  strategies.get('metrics')), results))

    # save results to folder
    filepaths = list(map(lambda x: os.path.join(save_dir, f'{x}.parquet'), range(len(modules))))
    list(map(lambda x: x[0].to_parquet(x[1], index=False), zip(results, filepaths)))  # execute save to parquet
    filenames = list(map(lambda x: os.path.basename(x), filepaths))

    # make summary file
    summary_df = pd.DataFrame({
        'filename': filenames,
        'module_name': list(map(lambda module: module.__name__, modules)),
        'module_params': module_params,
        'execution_time': average_times,
        **{f'passage_compressor_{metric}': list(map(lambda result: result[metric].mean(), results)) for metric in
           strategies.get('metrics')},
    })

    # filter by strategies
    if strategies.get('speed_threshold') is not None:
        results, filenames = filter_by_threshold(results, average_times, strategies['speed_threshold'], filenames)
    selected_result, selected_filename = select_best_average(results, strategies.get('metrics'), filenames)
    new_retrieved_contents = selected_result['retrieved_contents']
    previous_result['retrieved_contents'] = new_retrieved_contents
    selected_result = selected_result.drop(columns=['retrieved_contents'])
    best_result = pd.concat([previous_result, selected_result], axis=1)

    # add 'is_best' column to summary file
    summary_df['is_best'] = summary_df['filename'] == selected_filename

    # add prefix 'passage_compressor' to best_result columns
    best_result = best_result.rename(columns={
        metric_name: f'passage_compressor_{metric_name}' for metric_name in strategies.get('metrics')
    })

    # save the result files
    best_result.to_parquet(os.path.join(save_dir, f'best_{os.path.splitext(selected_filename)[0]}.parquet'),
                           index=False)
    summary_df.to_csv(os.path.join(save_dir, 'summary.csv'), index=False)
    return best_result


def evaluate_passage_compressor_node(result_df: pd.DataFrame,
                                     retrieval_contents_gt: List[List[str]],
                                     metrics: List[str]):
    metric_funcs = {
        retrieval_token_recall.__name__: retrieval_token_recall,
        retrieval_token_precision.__name__: retrieval_token_precision,
        retrieval_token_f1.__name__: retrieval_token_f1,
    }
    metrics = list(filter(lambda x: x in metric_funcs.keys(), metrics))
    if len(metrics) <= 0:
        raise ValueError(f"metrics must be one of {metric_funcs.keys()}")
    metrics_scores = dict(map(lambda metric: (metric, metric_funcs[metric](
        gt_contents=retrieval_contents_gt,
        pred_contents=result_df['retrieved_contents'].tolist()
    )), metrics))
    result_df = pd.concat([result_df, pd.DataFrame(metrics_scores)], axis=1)
    return result_df


def make_contents_gt(qa_data: pd.DataFrame, corpus_data: pd.DataFrame) -> List[List[str]]:
    validate_qa_dataset(qa_data)
    validate_corpus_dataset(corpus_data)
    retrieval_gt_ids = qa_data['retrieval_gt'].apply(lambda x: list(itertools.chain.from_iterable(x))).tolist()
    return fetch_contents(corpus_data, retrieval_gt_ids)



import glob
import json
import logging
import os
import shutil
from datetime import datetime
from itertools import chain
from typing import List, Dict, Optional

import chromadb
import pandas as pd
import torch
import yaml

from autorag import embedding_models
from autorag.node_line import run_node_line
from autorag.nodes.retrieval.bm25 import bm25_ingest
from autorag.nodes.retrieval.vectordb import vectordb_ingest
from autorag.schema import Node
from autorag.schema.node import module_type_exists, extract_values_from_nodes
from autorag.utils import cast_qa_dataset, cast_corpus_dataset, validate_qa_from_corpus_dataset
from autorag.utils.util import load_summary_file, convert_string_to_tuple_in_dict, convert_env_in_dict, explode

logger = logging.getLogger("AutoRAG")

ascii_art = """
                _        _____            _____ 
     /\        | |      |  __ \     /\   / ____|
    /  \  _   _| |_ ___ | |__) |   /  \ | |  __ 
   / /\ \| | | | __/ _ \|  _  /   / /\ \| | |_ |
  / ____ \ |_| | || (_) | | \ \  / ____ \ |__| |
 /_/    \_\__,_|\__\___/|_|  \_\/_/    \_\_____|
                                                
"""


class Evaluator:
    def __init__(self, qa_data_path: str, corpus_data_path: str, project_dir: Optional[str] = None):
        """
        Initialize an Evaluator object.

        :param qa_data_path: The path to the QA dataset.
            Must be parquet file.
        :param corpus_data_path: The path to the corpus dataset.
            Must be parquet file.
        :param project_dir: The path to the project directory.
            Default is the current directory.
        """
        # validate data paths
        if not os.path.exists(qa_data_path):
            raise ValueError(f"QA data path {qa_data_path} does not exist.")
        if not os.path.exists(corpus_data_path):
            raise ValueError(f"Corpus data path {corpus_data_path} does not exist.")
        if not qa_data_path.endswith('.parquet'):
            raise ValueError(f"QA data path {qa_data_path} is not a parquet file.")
        if not corpus_data_path.endswith('.parquet'):
            raise ValueError(f"Corpus data path {corpus_data_path} is not a parquet file.")
        self.qa_data = pd.read_parquet(qa_data_path)
        self.corpus_data = pd.read_parquet(corpus_data_path)
        self.qa_data = cast_qa_dataset(self.qa_data)
        self.corpus_data = cast_corpus_dataset(self.corpus_data)
        self.project_dir = project_dir if project_dir is not None else os.getcwd()
        if not os.path.exists(self.project_dir):
            os.makedirs(self.project_dir)

        validate_qa_from_corpus_dataset(self.qa_data, self.corpus_data)

        # copy dataset to project directory
        if not os.path.exists(os.path.join(self.project_dir, 'data')):
            os.makedirs(os.path.join(self.project_dir, 'data'))
        qa_path_in_project = os.path.join(self.project_dir, 'data', 'qa.parquet')
        if not os.path.exists(qa_path_in_project):
            self.qa_data.to_parquet(qa_path_in_project, index=False)
        corpus_path_in_project = os.path.join(self.project_dir, 'data', 'corpus.parquet')
        if not os.path.exists(corpus_path_in_project):
            self.corpus_data.to_parquet(corpus_path_in_project, index=False)

    def start_trial(self, yaml_path: str):
        logger.info(ascii_art)

        trial_name = self.__get_new_trial_name()
        self.__make_trial_dir(trial_name)

        # copy yaml file to trial directory
        shutil.copy(yaml_path, os.path.join(self.project_dir, trial_name, 'config.yaml'))
        node_lines = self._load_node_lines(yaml_path)
        self.__embed(node_lines)

        trial_summary_df = pd.DataFrame(columns=['node_line_name', 'node_type', 'best_module_filename',
                                                 'best_module_name', 'best_module_params', 'best_execution_time'])
        for i, (node_line_name, node_line) in enumerate(node_lines.items()):
            node_line_dir = os.path.join(self.project_dir, trial_name, node_line_name)
            os.makedirs(node_line_dir, exist_ok=False)
            if i == 0:
                previous_result = self.qa_data
            logger.info(f'Running node line {node_line_name}...')
            previous_result = run_node_line(node_line, node_line_dir, previous_result)

            trial_summary_df = self._append_node_line_summary(node_line_name, node_line_dir, trial_summary_df)

        trial_summary_df.to_csv(os.path.join(self.project_dir, trial_name, 'summary.csv'), index=False)

    def __embed(self, node_lines: Dict[str, List[Node]]):
        if any(list(map(lambda nodes: module_type_exists(nodes, 'bm25'), node_lines.values()))):
            # ingest BM25 corpus
            logger.info('Embedding BM25 corpus...')
            bm25_dir = os.path.join(self.project_dir, 'resources', 'bm25.pkl')
            if not os.path.exists(os.path.dirname(bm25_dir)):
                os.makedirs(os.path.dirname(bm25_dir))
            if os.path.exists(bm25_dir):
                logger.debug('BM25 corpus already exists.')
            else:
                bm25_ingest(bm25_dir, self.corpus_data)
            logger.info('BM25 corpus embedding complete.')
        if any(list(map(lambda nodes: module_type_exists(nodes, 'vectordb'), node_lines.values()))):
            # load embedding_models in nodes
            embedding_models_list = list(chain.from_iterable(
                map(lambda nodes: extract_values_from_nodes(nodes, 'embedding_model'), node_lines.values())))

            # duplicate check in embedding_models
            embedding_models_list = list(set(embedding_models_list))

            vectordb_dir = os.path.join(self.project_dir, 'resources', 'chroma')
            vectordb = chromadb.PersistentClient(path=vectordb_dir)

            for embedding_model_str in embedding_models_list:
                # ingest VectorDB corpus
                logger.info(f'Embedding VectorDB corpus with {embedding_model_str}...')

                # Get the collection with GET or CREATE, as it may already exist
                collection = vectordb.get_or_create_collection(name=embedding_model_str,
                                                               metadata={"hnsw:space": "cosine"})
                # get embedding_model
                if embedding_model_str in embedding_models:
                    embedding_model = embedding_models[embedding_model_str]
                else:
                    logger.error(f"embedding_model_str {embedding_model_str} does not exist.")
                    raise KeyError(f"embedding_model_str {embedding_model_str} does not exist.")
                vectordb_ingest(collection, self.corpus_data, embedding_model)
                logger.info(f'VectorDB corpus embedding complete with {embedding_model_str}.')
                del embedding_model
                if torch.cuda.is_available():
                    torch.cuda.empty_cache()
        else:
            logger.info('No ingestion needed.')

    def __get_new_trial_name(self) -> str:
        trial_json_path = os.path.join(self.project_dir, 'trial.json')
        if not os.path.exists(trial_json_path):
            return '0'
        with open(trial_json_path, 'r') as f:
            trial_json = json.load(f)
        return str(int(trial_json[-1]['trial_name']) + 1)

    def __make_trial_dir(self, trial_name: str):
        trial_json_path = os.path.join(self.project_dir, 'trial.json')
        if os.path.exists(trial_json_path):
            with open(trial_json_path, 'r') as f:
                trial_json = json.load(f)
        else:
            trial_json = []

        trial_json.append({
            'trial_name': trial_name,
            'start_time': datetime.now().strftime('%Y-%m-%d %H:%M:%S'),
        })
        os.makedirs(os.path.join(self.project_dir, trial_name))
        with open(trial_json_path, 'w') as f:
            json.dump(trial_json, f, indent=4)

    @staticmethod
    def _load_node_lines(yaml_path: str) -> Dict[str, List[Node]]:
        if not os.path.exists(yaml_path):
            raise ValueError(f"YAML file {yaml_path} does not exist.")
        with open(yaml_path, 'r') as stream:
            try:
                yaml_dict = yaml.safe_load(stream)
            except yaml.YAMLError as exc:
                raise ValueError(f"YAML file {yaml_path} could not be loaded.") from exc

        yaml_dict = convert_string_to_tuple_in_dict(yaml_dict)
        yaml_dict = convert_env_in_dict(yaml_dict)
        node_lines = yaml_dict['node_lines']
        node_line_dict = {}
        for node_line in node_lines:
            node_line_dict[node_line['node_line_name']] = list(
                map(lambda node: Node.from_dict(node), node_line['nodes']))
        return node_line_dict

    def restart_trial(self, trial_path: str):
        logger.info(ascii_art)
        # Check if trial_path exists
        if not os.path.exists(trial_path):
            raise ValueError(f"Trial path {trial_path} does not exist.")
        # Check if trial is completed
        if os.path.exists(os.path.join(trial_path, 'summary.csv')):
            raise ValueError(f"Trial path {trial_path} is already completed.")

        # Extract node lines from config.yaml
        yaml_path = os.path.join(trial_path, 'config.yaml')
        node_lines = self._load_node_lines(yaml_path)

        node_line_names = list(node_lines.keys())
        nodes = list(node_lines.values())
        node_names = list(map(lambda node: list(map(lambda n: n.node_type, node)), nodes))

        # If the First Node Line folder hasn't even been created, proceed to start_trial
        if not os.path.exists(os.path.join(trial_path, node_line_names[0])):
            self.start_trial(yaml_path)
            return None

        # Find conflict node line and node
        conflict_line_name, conflict_node_name = self.__find_conflict_point(trial_path, node_line_names, node_lines)
        node_dir = os.path.join(trial_path, conflict_line_name, conflict_node_name)
        if os.path.exists(node_dir):
            shutil.rmtree(node_dir)

        # Set remain_nodes and remain_lines
        remain_nodes, completed_node_names, remain_lines, remain_line_names \
            = self._set_remain_nodes_and_lines(node_line_names, nodes, node_names,
                                               conflict_node_name, conflict_line_name)
        # Set previous_result
        previous_result = self.__set_previous_result(node_line_names, node_names, trial_path, conflict_node_name)

        # Run Node
        if remain_nodes:
            conflict_line_dir = os.path.join(trial_path, conflict_line_name)
            summary_lst = []
            # Get already run node summary and append to summary_lst
            for completed_node_name in completed_node_names:
                summary_lst = self._append_node_summary(conflict_line_dir, completed_node_name, summary_lst)
            for node in remain_nodes:
                previous_result = node.run(previous_result, conflict_line_dir)
                summary_lst = self._append_node_summary(conflict_line_dir, node.node_type, summary_lst)
            pd.DataFrame(summary_lst).to_csv(os.path.join(conflict_line_dir, 'summary.csv'), index=False)

        # Run node line
        trial_summary_df = pd.DataFrame(columns=['node_line_name', 'node_type', 'best_module_filename',
                                                 'best_module_name', 'best_module_params', 'best_execution_time'])
        completed_line_names = node_line_names[:node_line_names.index(conflict_line_name)]
        # Get already run node line's summary and append to trial_summary_df
        if completed_line_names:
            for line_name in completed_line_names:
                node_line_dir = os.path.join(trial_path, line_name)
                trial_summary_df = self._append_node_line_summary(line_name, node_line_dir, trial_summary_df)
        if remain_lines:
            for node_line_name, node_line in zip(remain_line_names, remain_lines):
                node_line_dir = os.path.join(trial_path, node_line_name)
                if not os.path.exists(node_line_dir):
                    os.makedirs(node_line_dir)
                logger.info(f'Running node line {node_line_name}...')
                previous_result = run_node_line(node_line, node_line_dir, previous_result)
                trial_summary_df = self._append_node_line_summary(node_line_name, node_line_dir, trial_summary_df)
        trial_summary_df.to_csv(os.path.join(trial_path, 'summary.csv'), index=False)

    def __find_conflict_point(self, trial_path: str, node_line_names: List[str],
                              node_lines: Dict[str, List[Node]]) -> tuple[str, str]:
        for node_line_name in node_line_names:
            node_line_dir = os.path.join(trial_path, node_line_name)
            if not os.path.exists(node_line_dir):
                return node_line_name, node_lines[node_line_name][0].node_type

            if not os.path.exists(os.path.join(node_line_dir, 'summary.csv')):
                conflict_node_name = self._find_conflict_node_name(node_line_dir, node_lines[node_line_name])
                return node_line_name, conflict_node_name

        raise ValueError(f"No error node line found in {trial_path}.")

    @staticmethod
    def _find_conflict_node_name(node_line_dir: str, node_line: List[Node]) -> str:
        for node in node_line:
            node_dir = os.path.join(node_line_dir, node.node_type)
            if not os.path.exists(node_dir) or not os.path.exists(os.path.join(node_dir, 'summary.csv')):
                return node.node_type
        raise TypeError("No conflict node name found.")

    def __set_previous_result(self, node_line_names: List[str], node_names: List[List[str]],
                              trial_path: str, conflict_node_name: str):
        exploded_node_line, exploded_node = explode(node_line_names, node_names)
        conflict_node_index = exploded_node.index(conflict_node_name)
        # Set previous_result
        if conflict_node_index == 0:
            previous_result = self.qa_data
        else:
            previous_node_line = exploded_node_line[conflict_node_index - 1]
            previous_node = exploded_node[conflict_node_index - 1]

            previous_node_dir = os.path.join(trial_path, previous_node_line, previous_node)
            best_file_pattern = f'{previous_node_dir}/best_*.parquet'
            previous_result = pd.read_parquet(glob.glob(best_file_pattern)[0])
        return previous_result

    @staticmethod
    def _set_remain_nodes_and_lines(node_line_names: List[str], nodes: List[List[Node]],
                                    node_names: List[List[str]], conflict_node_name: str,
                                    conflict_node_line_name: str):
        conflict_node_line_index = node_line_names.index(conflict_node_line_name)
        full_conflict_node_line_nodes = nodes[conflict_node_line_index]
        full_conflict_node_line_node_names = node_names[conflict_node_line_index]

        if conflict_node_name == full_conflict_node_line_node_names[0]:
            remain_nodes = None
            completed_node_names = None
            remain_node_lines = nodes[conflict_node_line_index:]
            remain_node_line_names = node_line_names[conflict_node_line_index:]
        else:
            conflict_node_index = full_conflict_node_line_node_names.index(conflict_node_name)
            remain_nodes = full_conflict_node_line_nodes[conflict_node_index:]
            completed_node_names = full_conflict_node_line_node_names[:conflict_node_index]
            if conflict_node_line_index + 1 >= len(node_line_names):
                remain_node_lines = None
                remain_node_line_names = None
            else:
                remain_node_lines = nodes[conflict_node_line_index + 1:]
                remain_node_line_names = node_line_names[conflict_node_line_index + 1:]
        return remain_nodes, completed_node_names, remain_node_lines, remain_node_line_names

    @staticmethod
    def _append_node_line_summary(node_line_name: str, node_line_dir: str, trial_summary_df: pd.DataFrame):
        summary_df = load_summary_file(os.path.join(node_line_dir, 'summary.csv'),
                                       dict_columns=['best_module_params'])
        summary_df = summary_df.assign(node_line_name=node_line_name)
        summary_df = summary_df[list(trial_summary_df.columns)]
        if len(trial_summary_df) <= 0:
            trial_summary_df = summary_df
        else:
            trial_summary_df = pd.concat([trial_summary_df, summary_df], ignore_index=True)
        return trial_summary_df

    @staticmethod
    def _append_node_summary(node_line_dir: str, node_name: str, summary_lst: List[Dict]):
        node_summary_df = load_summary_file(os.path.join(node_line_dir, node_name, 'summary.csv'))
        best_node_row = node_summary_df.loc[node_summary_df['is_best']]
        summary_lst.append({
            'node_type': node_name,
            'best_module_filename': best_node_row['filename'].values[0],
            'best_module_name': best_node_row['module_name'].values[0],
            'best_module_params': best_node_row['module_params'].values[0],
            'best_execution_time': best_node_row['execution_time'].values[0],
        })
        return summary_lst



import asyncio
import itertools
import os
import pathlib
import tempfile
from datetime import datetime

import pandas as pd
import pytest
import tiktoken
from llama_index.core.llms import CompletionResponse

from autorag.utils import fetch_contents
from autorag.utils.util import load_summary_file, result_to_dataframe, \
    make_combinations, explode, replace_value_in_dict, normalize_string, convert_string_to_tuple_in_dict, process_batch, \
    convert_env_in_dict, openai_truncate_by_token
from tests.mock import MockLLM

root_dir = pathlib.PurePath(os.path.dirname(os.path.realpath(__file__))).parent.parent

summary_df = pd.DataFrame({
    'best_module_name': ['bm25', 'upr', 'gpt-4'],
    'best_module_params': [{'top_k': 50}, {'model': 'llama-2', 'havertz': 'chelsea'}, {'top_p': 0.9}],
})


@pytest.fixture
def module_name():
    return "test_module"


@pytest.fixture
def module_params():
    return {
        "param1": "value1",
        "param2": "value2",
        "param3": "value3",
    }


@pytest.fixture
def summary_path():
    with tempfile.TemporaryDirectory() as tmp_dir:
        summary_path = os.path.join(tmp_dir, "summary.csv")
        summary_df.to_csv(summary_path, index=False)
        yield summary_path


def test_fetch_contents():
    corpus_data_path = os.path.join(root_dir, "resources", "corpus_data_sample.parquet")
    corpus_data = pd.read_parquet(corpus_data_path)
    search_rows = corpus_data.sample(n=10)
    find_contents = fetch_contents(corpus_data, list(map(lambda x: [x], search_rows['doc_id'].tolist())))
    assert len(find_contents) == len(search_rows)
    assert list(itertools.chain.from_iterable(find_contents)) == search_rows['contents'].tolist()

    corpus_data = pd.DataFrame({
        'doc_id': ['doc1', 'doc2', 'doc3'],
        'contents': ['apple', 'banana', 'cherry'],
        'metadata': [{'last_modified_datetime': datetime(2022, 1, 1, 0, 0, 0)},
                     {'last_modified_datetime': datetime(2022, 1, 2, 0, 0, 0)},
                     {'last_modified_datetime': datetime(2022, 1, 3, 0, 0, 0)}]
    })
    find_contents = fetch_contents(corpus_data, [['doc3', 'doc1'], ['doc2']])
    assert find_contents[0] == ['cherry', 'apple']
    assert find_contents[1] == ['banana']

    find_metadatas = fetch_contents(corpus_data, [['doc3', 'doc1'], ['doc2']], 'metadata')
    assert find_metadatas[0] == [{'last_modified_datetime': datetime(2022, 1, 3, 0, 0, 0)},
                                 {'last_modified_datetime': datetime(2022, 1, 1, 0, 0, 0)}]
    assert find_metadatas[1] == [{'last_modified_datetime': datetime(2022, 1, 2, 0, 0, 0)}]

    find_empty = fetch_contents(corpus_data, [[], ['doc2']])
    assert find_empty[0] == [None]
    assert find_empty[1] == ['banana']

    find_blank = fetch_contents(corpus_data, [[''], ['doc2']])
    assert find_blank[0] == [None]
    assert find_blank[1] == ['banana']


def test_load_summary_file(summary_path):
    with pytest.raises(ValueError):
        load_summary_file(summary_path)
    df = load_summary_file(summary_path, ['best_module_params'])
    assert df.equals(summary_df)


def test_result_to_dataframe():
    @result_to_dataframe(['col_1', 'col_2'])
    def func1():
        return [1, 2], [3, 4]

    result1 = func1()
    assert isinstance(result1, pd.DataFrame)
    assert result1.columns.tolist() == ['col_1', 'col_2']
    assert result1['col_1'].tolist() == [1, 2]
    assert result1['col_2'].tolist() == [3, 4]

    @result_to_dataframe(['col_1'])
    def func2():
        return [1, 2, 3]

    result2 = func2()
    assert isinstance(result2, pd.DataFrame)
    assert result2.columns.tolist() == ['col_1']
    assert result2['col_1'].tolist() == [1, 2, 3]


def test_make_combinations():
    target_dict = {'key1': 'value1', 'key2': ['value1', 'value2'], 'key3': 'value3', 'key4': ['value4', 'value5']}
    solution = [
        {'key1': 'value1', 'key2': 'value1', 'key3': 'value3', 'key4': 'value4'},
        {'key1': 'value1', 'key2': 'value1', 'key3': 'value3', 'key4': 'value5'},
        {'key1': 'value1', 'key2': 'value2', 'key3': 'value3', 'key4': 'value4'},
        {'key1': 'value1', 'key2': 'value2', 'key3': 'value3', 'key4': 'value5'}
    ]
    combinations = make_combinations(target_dict)
    assert len(combinations) == len(solution)
    assert all([combination in solution for combination in combinations])

    elem1 = {
        'key5': 'value5',
        'key6': ['value6', 'value7']
    }
    elem2 = {'key7': 'value8'}
    value_of_key_4 = [elem1, elem2]
    target_dict = {'key1': 'value1', 'key2': ['value1', 'value2'], 'key3': 'value3', 'key4': value_of_key_4}
    combinations = make_combinations(target_dict)
    solution = [
        {'key1': 'value1', 'key2': 'value1', 'key3': 'value3', 'key4': elem1},
        {'key1': 'value1', 'key2': 'value2', 'key3': 'value3', 'key4': elem1},
        {'key1': 'value1', 'key2': 'value1', 'key3': 'value3', 'key4': elem2},
        {'key1': 'value1', 'key2': 'value2', 'key3': 'value3', 'key4': elem2},
    ]
    assert len(combinations) == len(solution)
    assert all([combination in solution for combination in combinations])

    target_dict = {'key1': 'value1', 'key2': ['value1', 'value2'], 'key3': 'value3', 'key4': ('value4', 'value5')}
    solution = [
        {'key1': 'value1', 'key2': 'value1', 'key3': 'value3', 'key4': ('value4', 'value5')},
        {'key1': 'value1', 'key2': 'value2', 'key3': 'value3', 'key4': ('value4', 'value5')},
    ]
    combinations = make_combinations(target_dict)
    assert len(combinations) == len(solution)
    assert all([combination in solution for combination in combinations])


def test_explode():
    index_values = ['a', 'b', 'c']
    explode_values = [
        ['apple', 'banana', 'cherry'],
        ['april', 'may'],
        ['alpha'],
    ]
    result_index, result_values = explode(index_values, explode_values)
    assert result_index == ['a', 'a', 'a', 'b', 'b', 'c']
    assert result_values == ['apple', 'banana', 'cherry', 'april', 'may', 'alpha']


def test_replace_value_in_dict():
    target_dict = {
        'key1': 'value1',
        'key2': 'value2',
        'key3': 'value3',
    }
    result_dict = replace_value_in_dict(target_dict, 'key1', 'value4')
    assert result_dict == {
        'key1': 'value4',
        'key2': 'value2',
        'key3': 'value3',
    }
    result_dict = replace_value_in_dict(target_dict, 'key4', 'value4')
    assert result_dict == target_dict


def test_normalize_string():
    text = "This IS a TEST Text."
    expected = "this is test text"
    assert normalize_string(text) == expected

    text = "Hello, world! This is a test."
    expected = "hello world this is test"
    assert normalize_string(text) == expected

    text = "The quick brown fox jumps over the lazy dog."
    expected = "quick brown fox jumps over lazy dog"
    assert normalize_string(text) == expected

    text = "This    is      a test    text."
    expected = "this is test text"
    assert normalize_string(text) == expected

    text = "The, QUICK Brown-Fox; jumps over... the LAZY dog!"
    expected = "quick brownfox jumps over lazy dog"
    assert normalize_string(text) == expected


def test_convert_string_to_tuple_in_dict():
    # Example usage
    data = {
        'key1': '(1, \'two\', 3)',
        'key2': ['(4, 5, \'six\')', {'nested_key': '(7, 8, \'nine\')'},
                 {'key4': 'value2'}],
        'key3': {'nested_key2': '(10, \'eleven\', 12)',
                 'nested_key3': 'value1',
                 'nested_key4': {'nested_key5': '(\'thirteen\', 14, 15)'}},
    }
    result = convert_string_to_tuple_in_dict(data)
    assert result == {
        'key1': (1, 'two', 3),
        'key2': [
            (4, 5, 'six'),
            {
                'nested_key': (7, 8, 'nine')
            },
            {
                'key4': 'value2'
            }
        ],
        'key3': {
            'nested_key2': (10, 'eleven', 12),
            'nested_key3': 'value1',
            'nested_key4': {
                'nested_key5': ('thirteen', 14, 15)
            }
        }
    }


def test_convert_env_in_dict():
    os.environ['ENV_VAR1'] = 'value1'
    os.environ['ENV_VAR2'] = 'value2'
    os.environ['ENV_VAR3'] = 'value3'
    data = {
        'key1': 'value1',
        'key2': [
            'value1',
            '${ENV_VAR1}',
        ],
        'key3': '${ENV_VAR2}',
        'key4': {
            'key5': 'value1',
            'key6': '${ENV_VAR3}',
            'key7': [
                'value1',
                '${ENV_VAR4}',
            ]
        },
        'prompt': 'This is a prompt with ${ENV_VAR1} and ${ENV_VAR2}.'
    }
    result = convert_env_in_dict(data)
    assert result == {
        'key1': 'value1',
        'key2': [
            'value1',
            'value1',
        ],
        'key3': 'value2',
        'key4': {
            'key5': 'value1',
            'key6': 'value3',
            'key7': [
                'value1',
                '',
            ]
        },
        'prompt': 'This is a prompt with value1 and value2.'
    }


def test_process_batch():
    prompts = [str(i) for i in range(1000)]
    results = [CompletionResponse(text=prompt) for prompt in prompts]
    mock_llm = MockLLM()

    tasks = [mock_llm.acomplete(prompt) for prompt in prompts]
    loop = asyncio.get_event_loop()
    result = loop.run_until_complete(process_batch(tasks, batch_size=64))

    assert result == results


def test_openai_truncate_by_token():
    base_text = "This is a test text."
    t1 = base_text * 5
    t2 = base_text * 2000
    t3 = base_text * 20

    truncated = openai_truncate_by_token([t1, t2, t3], 8192,
                                         'text-embedding-ada-002')
    assert len(truncated) == 3
    assert truncated[0] == base_text * 5
    assert len(truncated[1]) < len(t2)
    assert len(tiktoken.encoding_for_model('text-embedding-ada-002').encode(truncated[1])) == 8192
    assert len(truncated[2]) == len(t3)



from autorag.nodes.promptmaker import long_context_reorder
from tests.autorag.nodes.promptmaker.test_prompt_maker_base import (prompt, queries, retrieved_contents,
                                                                    retrieve_scores, previous_result)


def test_long_context_reorder():
    long_context_reorder_original = long_context_reorder.__wrapped__
    result_prompts = long_context_reorder_original(prompt, queries, retrieved_contents, retrieve_scores)
    assert len(result_prompts) == 2
    assert isinstance(result_prompts, list)
    assert result_prompts[
               0] == 'Answer this question: What is the capital of Japan? \n\n Tokyo is the capital of Japan.\n\nTokyo, the capital of Japan, is a huge metropolitan city.\n\nTokyo is the capital of Japan.'
    assert result_prompts[
               1] == 'Answer this question: What is the capital of China? \n\n Beijing is the capital of China.\n\nBeijing, the capital of China, is a huge metropolitan city.\n\nBeijing is the capital of China.'


def test_long_context_reorder_node():
    result = long_context_reorder(project_dir="pseudo_project_dir",
                                  previous_result=previous_result,
                                  prompt=prompt)
    assert len(result) == 2
    assert result.columns == ["prompts"]
    assert result['prompts'][
               0] == "Answer this question: What is the capital of Japan? \n\n Tokyo is the capital of Japan.\n\nTokyo, the capital of Japan, is a huge metropolitan city.\n\nTokyo is the capital of Japan."
    assert result['prompts'][
               1] == "Answer this question: What is the capital of China? \n\n Beijing is the capital of China.\n\nBeijing, the capital of China, is a huge metropolitan city.\n\nBeijing is the capital of China."



import itertools

import pytest
import torch
from transformers import AutoModel, AutoTokenizer

from autorag.nodes.passagereranker import colbert_reranker
from autorag.nodes.passagereranker.colbert import get_colbert_embedding_batch, slice_tensor
from tests.autorag.nodes.passagereranker.test_passage_reranker_base import queries_example, contents_example, \
    scores_example, ids_example, base_reranker_test, project_dir, previous_result, base_reranker_node_test
from tests.delete_tests import is_github_action


@pytest.mark.skipif(is_github_action(), reason="Skipping this test on GitHub Actions because it uses local model.")
def test_colbert_reranker():
    top_k = 2
    original_colbert_reranker = colbert_reranker.__wrapped__
    contents_result, id_result, score_result \
        = original_colbert_reranker(queries_example, contents_example, scores_example, ids_example, top_k)
    base_reranker_test(contents_result, id_result, score_result, top_k)


@pytest.mark.skipif(is_github_action(), reason="Skipping this test on GitHub Actions because it uses local model.")
def test_colbert_reranker_long():
    top_k = 2
    original_colbert_reranker = colbert_reranker.__wrapped__
    contents_example[0][0] = contents_example[0][0] * 10000
    contents_result, id_result, score_result \
        = original_colbert_reranker(queries_example, contents_example, scores_example, ids_example, top_k, batch=2)
    base_reranker_test(contents_result, id_result, score_result, top_k)


@pytest.mark.skipif(is_github_action(), reason="Skipping this test on GitHub Actions because it uses local model.")
def test_colbert_reranker_one_batch():
    top_k = 2
    original_colbert_reranker = colbert_reranker.__wrapped__
    contents_result, id_result, score_result \
        = original_colbert_reranker(queries_example, contents_example, scores_example, ids_example, top_k, batch=1)
    base_reranker_test(contents_result, id_result, score_result, top_k)


@pytest.mark.skipif(is_github_action(), reason="Skipping this test on GitHub Actions because it uses local model.")
def test_colbert_reranker_node():
    top_k = 1
    result_df = colbert_reranker(project_dir=project_dir, previous_result=previous_result, top_k=top_k)
    base_reranker_node_test(result_df, top_k)


@pytest.mark.skipif(is_github_action(), reason="Skipping this test on GitHub Actions because it uses local model.")
def test_colbert_embedding():
    contents = list(itertools.chain.from_iterable(contents_example))
    model_name = "colbert-ir/colbertv2.0"
    device = "cuda" if torch.cuda.is_available() else "cpu"
    model = AutoModel.from_pretrained(model_name).to(device)
    tokenizer = AutoTokenizer.from_pretrained(model_name)
    colbert_embedding = get_colbert_embedding_batch(contents, model, tokenizer, batch_size=2)

    assert isinstance(colbert_embedding, list)
    assert len(colbert_embedding) == len(contents)
    assert colbert_embedding[0].shape == (1, 11, 768)


def test_slice_tensor():
    original_tensor = torch.randn(14, 7)
    batch_size = 4
    resulting_tensor_list = slice_tensor(original_tensor, batch_size)
    assert len(resulting_tensor_list) == 4
    assert resulting_tensor_list[0].size() == torch.Size([4, 7])
    assert resulting_tensor_list[-1].size() == torch.Size([2, 7])



import functools
from pathlib import Path
from typing import Union, Tuple, List

import pandas as pd

from autorag import generator_models
from autorag.utils import result_to_dataframe


def generator_node(func):
    @functools.wraps(func)
    @result_to_dataframe(["generated_texts", "generated_tokens", "generated_log_probs"])
    def wrapper(
            project_dir: Union[str, Path],
            previous_result: pd.DataFrame,
            llm: str,
            **kwargs) -> Tuple[List[str], List[List[int]], List[List[float]]]:
        """
        This decorator makes a generator module to be a node.
        It automatically extracts prompts from previous_result and runs the generator function.
        Plus, it retrieves the llm instance from autorag.generator_models.
        
        :param project_dir: The project directory.
        :param previous_result: The previous result that contains prompts,
        :param llm: The llm name that you want to use.
        :param kwargs: The extra parameters for initializing the llm instance.
        :return: Pandas dataframe that contains generated texts, generated tokens, and generated log probs.
            Each column is "generated_texts", "generated_tokens", and "generated_log_probs".
        """
        assert 'prompts' in previous_result.columns, "previous_result must contain prompts column."
        prompts = previous_result['prompts'].tolist()
        if func.__name__ == 'llama_index_llm':
            if llm not in generator_models:
                raise ValueError(f"{llm} is not a valid llm name. Please check the llm name."
                                 "You can check valid llm names from autorag.generator_models.")
            batch = kwargs.pop('batch', 16)
            llm_instance = generator_models[llm](**kwargs)
            result = func(prompts=prompts, llm=llm_instance, batch=batch)
            del llm_instance
            return result
        elif func.__name__ == 'vllm':
            return func(prompts=prompts, llm=llm, **kwargs)
        else:
            raise ValueError(f"{func.__name__} is not a valid generator node name. "
                             "Please check the generator node name.")

    return wrapper



from typing import List

from autorag.nodes.passagereranker.base import passage_reranker_node


@passage_reranker_node
def pass_reranker(queries: List[str], contents_list: List[List[str]],
                  scores_list: List[List[float]], ids_list: List[List[str]],
                  top_k: int):
    """
    Do not perform reranking.
    Return the given top-k passages as is.
    """
    contents_list = list(map(lambda x: x[:top_k], contents_list))
    scores_list = list(map(lambda x: x[:top_k], scores_list))
    ids_list = list(map(lambda x: x[:top_k], ids_list))
    return contents_list, ids_list, scores_list



import functools
from pathlib import Path
from typing import List, Union, Dict

import pandas as pd
from llama_index.core.service_context_elements.llm_predictor import LLMPredictorType

from autorag import generator_models
from autorag.utils import result_to_dataframe


def passage_compressor_node(func):
    @functools.wraps(func)
    @result_to_dataframe(['retrieved_contents'])
    def wrapper(
            project_dir: Union[str, Path],
            previous_result: pd.DataFrame,
            *args, **kwargs) -> List[List[str]]:
        assert all([column in previous_result.columns for column in
                    ['query', 'retrieved_contents', 'retrieved_ids', 'retrieve_scores']]), \
            "previous_result must have retrieved_contents, retrieved_ids, and retrieve_scores columns."
        assert len(previous_result) > 0, "previous_result must have at least one row."

        queries = previous_result['query'].tolist()
        retrieved_contents = previous_result['retrieved_contents'].tolist()
        retrieved_ids = previous_result['retrieved_ids'].tolist()
        retrieve_scores = previous_result['retrieve_scores'].tolist()

        if func.__name__ in ['tree_summarize', 'refine']:
            param_list = ['prompt', 'chat_prompt', 'context_window', 'num_output', 'batch']
            param_dict = dict(filter(lambda x: x[0] in param_list, kwargs.items()))
            kwargs_dict = dict(filter(lambda x: x[0] not in param_list, kwargs.items()))
            llm_name = kwargs_dict.pop('llm')
            llm = make_llm(llm_name, kwargs_dict)
            result = func(
                queries=queries,
                contents=retrieved_contents,
                scores=retrieve_scores,
                ids=retrieved_ids,
                llm=llm,
                **param_dict
            )
            del llm
            result = list(map(lambda x: [x], result))
        elif func.__name__ == 'pass_compressor':
            result = func(contents=retrieved_contents)
        else:
            raise ValueError(f"{func.__name__} is not supported in passage compressor node.")

        return result

    return wrapper


def make_llm(llm_name: str, kwargs: Dict) -> LLMPredictorType:
    if llm_name not in generator_models:
        raise KeyError(f"{llm_name} is not supported. "
                       "You can add it manually by calling autorag.generator_models.")
    return generator_models[llm_name](**kwargs)
